"""
Implements a widget for logging in to a server
"""


from PySide6 import QtCore, QtWidgets

from configurun.app.ui.network_login_widget_ui import Ui_NetworkLoginWidget



class NetworkLoginWidget(QtWidgets.QWidget):
	"""A widget to control connection to a server:
		Input-fields: IP, port, password
		Buttons: connect, disconnect, cancel
	"""
	#A controller to manage the machine learning window
	disconnectClicked = QtCore.Signal()
	connectClicked = QtCore.Signal(str, str, str) #IP, port, password
	cancelClicked = QtCore.Signal()


	def __init__(self, widget : QtWidgets.QWidget, settings : QtCore.QSettings) -> None:
		super().__init__()
		self._ui = Ui_NetworkLoginWidget()
		self._ui.setupUi(widget)
		self._settings = settings
		self.server_ip_history : list[str] = self._settings.value("server_ip_history", None) # type: ignore
		self.server_port_history : list[str] = self._settings.value("server_port_history", None) # type: ignore
		if self.server_ip_history is None or not isinstance(self.server_ip_history, list):
			self.server_ip_history = []
		if self.server_port_history is None or not isinstance(self.server_port_history, list):
			self.server_port_history = []
		assert isinstance(self.server_ip_history, list)
		assert isinstance(self.server_port_history, list)

		self._ui.serverIPComboBox.addItems(self.server_ip_history)
		self._ui.serverPortComboBox.addItems(self.server_port_history)

		#Set password-input to be a password input
		self._ui.serverPasswordLineEdit.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)

		self._history_max_count = 10
		self._ui.disconnectBtn.clicked.connect(self.disconnectClicked)
		self._ui.connectBtn.clicked.connect(lambda : self.connectClicked.emit(
			self._ui.serverIPComboBox.lineEdit().text(),
			self._ui.serverPortComboBox.lineEdit().text(),
			self._ui.serverPasswordLineEdit.text()
		))

		#Updates the history of the combobox (keeps max_count in account as well as duplicates)
		self._ui.serverIPComboBox.lineEdit().editingFinished.connect(
			lambda : self.combo_box_txt_changed(self._ui.serverIPComboBox, self.server_ip_history))
		self._ui.serverPortComboBox.lineEdit().editingFinished.connect(
			lambda : self.combo_box_txt_changed(self._ui.serverPortComboBox, self.server_port_history))

		self._ui.cancelBtn.clicked.connect(self.cancelClicked)

	#Pyqt slot that accept a triplet of strings (ip, port, password)
	@QtCore.Slot(str, str, str)
	def client_connected(self, server_ip : str, server_port : str, server_password : str) -> None:
		"""Disable the connect button and inputs and enable the disconnect button"""

		self._ui.serverIPComboBox.lineEdit().setText(server_ip)
		self._ui.serverPortComboBox.lineEdit().setText(server_port)
		self._ui.serverPasswordLineEdit.setText(server_password)

		self._ui.serverIPComboBox.setEnabled(False)
		self._ui.serverPortComboBox.setEnabled(False)
		self._ui.serverPasswordLineEdit.setEnabled(False)

		self._ui.connectBtn.setEnabled(False)
		self._ui.disconnectBtn.setEnabled(True)
		self._ui.cancelBtn.setDefault(True)

	@QtCore.Slot()
	def client_disconnected(self) -> None:
		"""Disable the disconnect button and enable the connect button and inputs"""
		self._ui.connectBtn.setEnabled(True)
		self._ui.disconnectBtn.setEnabled(False)

		self._ui.serverIPComboBox.setEnabled(True)
		self._ui.serverPortComboBox.setEnabled(True)
		self._ui.serverPasswordLineEdit.setEnabled(True)

		self._ui.serverIPComboBox.lineEdit().setText(
			self.server_ip_history[0] if len(self.server_ip_history) > 0 else ""
		)
		self._ui.serverPortComboBox.lineEdit().setText(
			self.server_port_history[0] if len(self.server_port_history) > 0 else ""
		)
		#Set connect button to be the default button
		self._ui.connectBtn.setDefault(True)


	def save_histories(self) -> None:
		"""Save the history of the comboboxes to the settings"""
		if len(self.server_ip_history) == 1: #TODO: somehow loading settings goes wrong if only 1 item in list
			self.server_ip_history.append('')
		if len(self.server_port_history) == 1:
			self.server_port_history.append('')

		self._settings.setValue("server_ip_history", self.server_ip_history)
		self._settings.setValue("server_port_history", self.server_port_history)

	def combo_box_txt_changed(self, combobox : QtWidgets.QComboBox, history_list : list):
		"""When the text in a combobox is changed, update the history list and the combobox

		Args:
			combobox (QtWidgets.QComboBox): The combobox that was changed
			history_list (list): The history list to update
		"""
		text = combobox.lineEdit().text()
		if text in history_list:
			history_list.remove(text)
		history_list.insert(0, text)
		if len(history_list) > self._history_max_count:
			history_list.pop(-1)

		combobox.clear()
		for address in history_list:
			combobox.addItem(address)


if __name__ == "__main__":
	import sys
	app = QtWidgets.QApplication(sys.argv)
	test_window = QtWidgets.QMainWindow()
	test_widget = QtWidgets.QWidget()
	loging_widget = NetworkLoginWidget(test_widget, QtCore.QSettings())
	loging_widget.client_connected("connectedip", "connectedport", "connectedpassword")
	test_window.setCentralWidget(test_widget)
	test_window.show()
	sys.exit(app.exec())



from .example_configuration import example_target_function
from .example_configuration import run_example_app
from .example_argparse import run_argparse_example
from .example_configuration import example_deduce_new_option_classes


""" 
Implements the base-options from which all user-defined options should inherit
"""


import logging
from dataclasses import dataclass
from pyside6_utils.classes import Serializable #Absolute import to avoid import PySide6
log = logging.getLogger(__name__)


@dataclass
class BaseOptions(Serializable, object):
	"""The base-option class, all other options classes should inherit from this class
	A configuration is built by using 1 or more of these options classes.
	"""

	def update_using_dict(self, update_dict : dict):
		"""Update the current object with the passed dictionary

		Args:
			update_dict (dict): The dictionary to update with
		"""
		#TODO: only set existing keys?
		for key, value in update_dict.items():
			setattr(self, key, value)


	def copy_from(self, other : 'BaseOptions'):
		"""Copy all values from other to self

		Args:
			other (BaseOptions): The other object to copy from
		"""
		for key, value in other.__dict__.items():
			setattr(self, key, value)

	def get_public_attrs_as_str(self, sep = "\n") -> str:
		"""Get all public attributes of the class as a string
		"""
		retval : str = ""
		for key, value in self.__dict__.items():
			if not key.startswith('_'):
				retval += f"{key} : {value}{sep}"
		return retval


	def __getitem__(self, key : str):
		"""Interface for []-itemgetters used in MVTS datasets, if key (class property) not found, returns None

		Args:
			key (str): The key of the item

		Returns:
			(typing.Any) : None if key not found, or the item if key found
		"""
		if key in self.__dict__:
			return self.__dict__[key]
		else:
			return None

	def __setitem__(self, key, value):
		"""Set the value of a property.

		Args:
			key (str): Name of the property
			value (any): Value to be set
		"""
		setattr(self, key, value)



"""
Implements a RunQueue model that synchronizes with text-output from running items in the runqueue.
"""


import copy
import datetime
import logging
import threading
import time
import typing
from collections import OrderedDict

from PySide6 import QtCore, QtGui, QtWidgets
from pyside6_utils.models.console_widget_models.console_model import BaseConsoleItem
from configurun.classes.run_queue import RunQueue

log = logging.getLogger(__name__)

class RunQueueConsoleItem(BaseConsoleItem):
	"""
	Model that synchronizes with text-output from running items in the runqueue.
	"""
	currentTextChanged = QtCore.Signal(str) #Emitted when the text in the file changes
	dataChanged = QtCore.Signal() #When the metadata of the item changes (e.g. last-edit-date, name, runState not txt)
	filledTextPositionsChanged = QtCore.Signal(object) #Emitted when the receieved text is filled in the buffer

	def __init__(self,
	    	item_id :int,
			name : str,
			path : str | None = None,
			active_state : bool = True
			):
		super().__init__()
		self._edit_mutex = threading.Lock() #When adding from/to the buffer, lock this mutex
		self._current_text : str = ""
		self._name : str = name
		self._last_edited : datetime.datetime | None = None
		self._id : int = item_id
		self._path : str | None = path
		# max_buffer : int = 100_000
		self._current_gap : list[int] = [0, 0] #We allow for 1 gap in the text

		self._filled_filepositions : list[tuple[int,int]]= [(0,0)] #A list of tuples (min, max) of filled filepositions
		 #If this is length 1, that means we have processed all data up to filled_filepositions[0][1] without gaps
		 # E.g. [(0, 100), (200, 300)] means we have 'processed' all chars from 0->99 and 200->299, but not 100->199

		self.runqueue_console_connection = None

		self._active_state = active_state #By default - assume that the item is running

		self._console_icon = QtGui.QIcon()
		self._console_icon.addFile(":/Icons/icons/actions/Console.png",
	       QtCore.QSize(),
		   QtGui.QIcon.Mode.Normal,
		   QtGui.QIcon.State.Off)

		self._done_console_icon = QtGui.QIcon()
		self._done_console_icon.addFile(":/Icons/icons/actions/Console checked.png",
	       QtCore.QSize(),
		   QtGui.QIcon.Mode.Normal,
		   QtGui.QIcon.State.Off)

	def get_id(self) -> int:
		"""Returns the id of this item - id is set at creation and should be unique for each item"""
		return self._id

	def set_active_state(self, new_active_state : bool):
		"""Sets the active state of the item (e.g. whether it is running or not)

		Args:
			new_active_state (bool): The new state
		"""
		self._active_state = new_active_state
		self.dataChanged.emit()

	def get_active_state(self) -> bool:
		"""Returns the active state of the item (e.g. whether it is running or not)

		Returns:
			bool: The active state of the item
		"""
		return self._active_state

	# def set_runqueue(self, run_queue : RunQueue):
	# 	self._run_queue = run_queue
	# 	if self.runqueue_console_connection is not None:
	# 		self.runqueue_console_connection.disconnect()
	# 	# self.reset()
	# 	#TODO: centralize the commandLineOutput into the ConsoleItemModel
	# 	self.runqueue_console_connection = self._run_queue.newCommandLineOutput.connect(self.on_commandline_output)



		# self.on_commandline_output(self._id, self._name, self._path, dt, 0, text)

	def on_commandline_output(self,
			   					item_id : int,
								name : str,
								output_path : str,
								edit_dt : datetime.datetime,
								filepos : int,
								msg : str,
								emit_datachanged : bool = True 
							): # pylint: disable=unused-argument
		"""Called when new command line output is received, inserts the new text into the buffer and emits 
		the currentTextChanged signal if the buffer changed
		
		Args:
			item_id (int): The id of the item for which the output is received
			name (str): The name of the item
			output_path (str): The path of the item
			edit_dt (datetime.datetime): The last (known) change to the output file
			filepos (int): The start-position where the new text should be inserted
			msg (str): The new text message
			emit_datachanged (bool, optional): Whether to emit the dataChanged signal if metadata changed. 
				Defaults to True. Set to false when using this function when resetting a model that uses this item
				to avoid emitting the dataChanged signal before the data of the model is fully reset.
		
		"""
		if item_id != self._id: #Skip if not the correct id
			return
		if name != self._name:
			raise ValueError(f"Name mismatch: {name} != {self._name} but id is the same - should not happen")

		data_changed = False #Whether the node-data itself changed (e.g. name (namechange not implemented) or edit_dt

		with self._edit_mutex:
			if self._last_edited is None or self._last_edited < edit_dt:
				self._last_edited = edit_dt
				data_changed = True

			filled_text_positions_changed = False

			cur_filepositions = copy.copy(self._filled_filepositions)

			#========== Insert new filled filepositions based on new msg ===========
			for i, (cur_min, cur_max) in enumerate(cur_filepositions):
				if filepos >= cur_min and filepos <= cur_max: #If filling from this interval
					if filepos + len(msg) <= cur_max: #If already filled this part
						break
					else:
						self._filled_filepositions[i] = (cur_min, filepos + len(msg))
						filled_text_positions_changed = True
						break
				elif i+1 < len(cur_filepositions): #If the next item is not the last item
					if filepos >= cur_max and filepos + len(msg) <= cur_filepositions[i+1][0]: #If filling in between 2
						self._filled_filepositions.insert(i+1, (filepos, filepos + len(msg)))
						filled_text_positions_changed = True
						break
				else: #If the last item, and filepos > max (already checked above)
					self._filled_filepositions.append((filepos, filepos + len(msg)))
					filled_text_positions_changed = True

			#===== Iterate over all filled filepositions and merge overlapping ones=======
			if filled_text_positions_changed:
				i = 0
				while i < (len(self._filled_filepositions)-1):
					if self._filled_filepositions[i][1] >= self._filled_filepositions[i+1][0]:
						# new_filled.append((cur_filepositions[i][0], cur_filepositions[i+1][1]))
						self._filled_filepositions[i] = (
							cur_filepositions[i][0], max(cur_filepositions[i+1][1], cur_filepositions[i][1]))
						del self._filled_filepositions[i+1]
						continue
					i+=1
			if filled_text_positions_changed:
				self.filledTextPositionsChanged.emit(self._filled_filepositions)

			#=========== Insert actual text into the buffer ===========
			#TODO: make use of a ringbuffer of user-specified size
			new_text = self._current_text[:filepos] if len(self._current_text) > filepos else \
				self._current_text + ("X" * (filepos - len(self._current_text)))
			new_text += msg
			self._current_text = new_text +(self._current_text[filepos + len(msg):] \
				if len(self._current_text) > filepos else "")


		self.currentTextChanged.emit(self._current_text)
		if data_changed and emit_datachanged: #If the metadata changed, emit the dataChanged signal
			self.dataChanged.emit()
			# if filepos + len(msg) > len(self._current_text):

	def get_current_text(self) -> str:
		return self._current_text

	def data(self, role : QtCore.Qt.ItemDataRole, column : int = 0):
		if role == QtCore.Qt.ItemDataRole.DisplayRole:
			if column == 0 :
				return f"{self._id}: {self._name}"
			elif column == 1 :
				return self._last_edited.strftime("%Y-%m-%d %H:%M:%S") if self._last_edited is not None else ""
			elif column == 2 :
				return self._path
			else:
				raise ValueError(f"Invalid column for ConsoleStandardItem (max=2 but is {column})")
		elif role == QtCore.Qt.ItemDataRole.EditRole:
			if column == 0 :
				return f"{self._id}: {self._name}"
			elif column == 1 :
				return self._last_edited
			elif column == 2 :
				return self._path
			else:
				raise ValueError(f"Invalid column for ConsoleStandardItem (max=2 but is {column})")
		# elif role == QtCore.Qt.ItemDataRole.DecorationRole:
		# 	return self._id
		elif role == QtCore.Qt.ItemDataRole.DecorationRole and column == 0: #if first column -> return active/inactive icon
			return self._console_icon if self._active_state else self._done_console_icon
		else:
			return None


class RunQueueConsoleModel(QtCore.QAbstractItemModel):
	"""
	A RunQueue model that synchronizes with text-output from running items in the runqueue by requesting all text-data
	from the runqueue upon connection, and then subscribing to any changes to the text-output of the items
	running in the runqueue.

	On reset, will try to fetch the history of the consoles using the runqueue fetch methods.
	"""


	def __init__(self, 
	    		parent: QtWidgets.QWidget | None = None,
				max_initial_console_history : int = 200_000
			) -> None:
		"""Initializes the runqueueconsolemodel

		Args:
			parent (QtWidgets.QWidget | None, optional): The parent. Defaults to None.
			max_initial_console_history (int, optional): Limits the max amount of characters that the model will 
				fetch on-reset (per item) this makes sure that the initial load times on model-reset do not become 
				too outrageous by only loading part of the data. Especially useful useful when working over network 
				connections. NOTE: this limit is not enforced when gradually loading data from the runqueue as 
				we assume that it's not feasible to load too much text-data in a single session.
				Defaults to 200_000. -1 for no limit.
		"""

		super().__init__(parent)
		self._run_queue = None
		self._id_item_map_mutex = threading.Lock()
		self._id_item_map : typing.OrderedDict[int, RunQueueConsoleItem] = OrderedDict({}) #Maps run queue id to item
			# Note: we use an ordered dict to keep a consistent item-order for the UI
		self._ignored_ids : typing.Set[int] = set() #Ids that are ignored/not tracked

		self._max_initial_console_history = max_initial_console_history

		self._new_cmd_text_signal : QtCore.SignalInstance | None= None
		self._active_ids_signal : QtCore.SignalInstance | None = None
		self._queue_reset_signal : QtCore.SignalInstance | None = None

	def set_run_queue(self, run_queue : RunQueue):
		"""Set the runqueue to synchronize with

		Args:
			run_queue (RunQueue): The new runqueue to synchronize with
		"""
		self._run_queue = run_queue
		if self._new_cmd_text_signal is not None:
			self._new_cmd_text_signal.disconnect()
		if self._active_ids_signal is not None:
			self._active_ids_signal.disconnect()
		if self._queue_reset_signal is not None:
			self._queue_reset_signal.disconnect()

		self._active_ids_signal = self._run_queue.currentlyRunningIdsChanged.connect(self.running_ids_changed)
		self._new_cmd_text_signal = self._run_queue.newCommandLineOutput.connect(self.new_command_line_output)
		self._queue_reset_signal = self._run_queue.resetTriggered.connect(self.reset) #Upon queue reset ->
			# re-request all data from the runqueue

	def reset(self, reset_ignored_ids : bool = False):
		"""
		Re-request all data from runQueue and fill model
		#TODO: maybe create a soft-reset that only checks if there is any missing data e.g. when reconnecting to
		a disconnected runqueue - not all data is neccesarily invalid
		"""
		log.info("Resetting RunQueueConsoleModel - re-requesting all console-output from runQueue")
		self.beginResetModel()
		with self._id_item_map_mutex:
			self._id_item_map = OrderedDict({})
		if reset_ignored_ids:
			self._ignored_ids = set()

		if self._run_queue is not None:
			cur_queue = self._run_queue.get_command_line_info_list()
			if cur_queue:
				for cur_id, (name, path, file_size, running) in cur_queue.items():
					if cur_id in self._ignored_ids:
						continue
					item = RunQueueConsoleItem(cur_id, name, path, running)
					self._append_row(item)

					if self._max_initial_console_history > 0 and file_size > self._max_initial_console_history: #if limit to initial console history
						all_txt, last_edit_dt = self._run_queue.get_command_line_output(
							cur_id, -1, self._max_initial_console_history) #TODO: this probably goes wrong if we resume
							# logging on an existing logfile, as we will "start again" at filepos 0
							# while the actual filepos = file_size-self._max_initial_console_history.
							# probably best to put a start_pos inside run_queue_item 
					else: #Fetch all data NOTE: this can be very costly, especially over network connections
						all_txt, last_edit_dt = self._run_queue.get_command_line_output(cur_id, -1, file_size)
					item.on_commandline_output( #Append all text to the item #TODO: do this before reading file to avoid
							# missing any of the data that was added during reading
						item_id=cur_id,
						name=name,
						output_path=path,
						edit_dt=last_edit_dt,
						filepos=0,
						msg=all_txt,
						emit_datachanged=False #Don't emit changes, otherwise we will try to update items that are not
							# yet known to the model
					) #TODO: maybe only import active items?



		self.endResetModel()
		log.info("Done resetting RunQueueConsoleModel")


	def get_ignored_ids(self) -> typing.List[int]:
		"""
		Get the ids that are ignored/not tracked.
		Returns:
			list[int]: A list of ids that are ignored/not tracked
		"""
		return list(self._ignored_ids)

	def new_command_line_output(self,
			  				item_id : int,
							item_name : str,
							item_path : str,
							item_edit_dt : datetime.datetime,
							item_filepos : int,
							item_msg : str):
		"""alled when new command line output is received

		Args:
			item_id (int): The id of the item for which the output is received
			item_name (str): The name of the item
			item_path (str): The path of the item
			item_edit_dt (datetime.datetime): The last (known) change to the output file
			item_filepos (int): The start-position where the new text should be inserted
			item_msg (str): The new text message
		"""
		if item_id in self._ignored_ids:
			return
		if item_id not in self._id_item_map:
			item = RunQueueConsoleItem(item_id, item_name, item_path)
			self.append_row(item)

		self._id_item_map[item_id].on_commandline_output(item_id, item_name, item_path, item_edit_dt, item_filepos, item_msg)

	def running_ids_changed(self, runnings_ids : typing.List[int]):
		"""Called when the running ids changed, updates the console-items to reflect the active/inactive state"""
		with self._id_item_map_mutex:
			for cur_id in self._id_item_map:
				self._id_item_map[cur_id].set_active_state(cur_id in runnings_ids)


	def columnCount(self, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> int: #pylint: disable=unused-argument
		return 3

	def rowCount(self, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> int: # pylint: disable=unused-argument
		return len(self._id_item_map)

	def un_ignore_id(self, item_id : int):
		"""Un-ignore the item with the passed id. Re-synchronizes text-data with the runqueue and re-adds the item
		to the model for the view to display
		TODO: We request the whole text file, but this might take a long while for very large files. Instead
		we should only request the missing data (e.g. the last 1000 chars) and append that to the current text
		Args:
			item_id (int): The item to un-ignore
		"""
		log.debug(f"Un-ignoring id {item_id}")
		with self._id_item_map_mutex:
			if item_id not in self._ignored_ids:
				return #Already un-ignored
			self._ignored_ids.remove(item_id)

		if self._run_queue is not None:
			cur_queue = self._run_queue.get_command_line_info_list()
			if item_id in cur_queue:
				name, path, file_size, currently_running = cur_queue[item_id]
				item = RunQueueConsoleItem(item_id, name, path, currently_running)
				self.append_row(item)
				self.dataChanged.emit(self.index(0,0), self.index(self.rowCount()-1, self.columnCount()-1))

				#TODO: instead of re-requesting all data, only request the missing data - don't delete upon ignore,
				#instead, simply don't show/update the item in the model
				all_txt, last_edit_dt = self._run_queue.get_command_line_output(item_id, -1, file_size)
				item.on_commandline_output( #Append all text to the item
					item_id=item_id,
					name=name,
					output_path=path,
					edit_dt=last_edit_dt,
					filepos=0,
					msg=all_txt
				) #TODO: maybe only import active items?

		# item = RunQueueConsoleItem()

	# def reset(self):
	# 	"""Re-request all data and refill model"""
	# 	self.beginResetModel()
	# 	self._id_item_map = {}
	# 	if self._run_queue is not None:
	# 		cur_queue = self._run_queue.get_command_line_info_list()

	# 		for id, (name, path, file_size) in cur_queue.items():
	# 			if id in self._ignored_ids:
	# 				continue
	# 			item = RunQueueConsoleItem(id, name, path)
	# 			with self._id_item_map_mutex:
	# 				self._id_item_map[id] = item
	# 			self.appendRow(item)
	# 	self.endResetModel()

	def _append_row(self, item : RunQueueConsoleItem):
		"""Append a row to the model - consisting of a single ConsoleStandardItem.
		NOTE: this is the same as appendRow, but without the begin/endInsertRows calls, better to be used during model
		reset to avoid emitting dataChanged/rowInserted signals for new items as this might cause issues when 
		begin/endresetModel is called during a reset
		
		Args:
			item (ConsoleStandardItem): The item to append (each row corresponds to 1 item)
		"""
		cur_item_id = item.get_id()
		with self._id_item_map_mutex:
			assert cur_item_id not in self._id_item_map, "Item with this ID already in model"
			self._id_item_map[cur_item_id] = item
			#TODO: only emit dataChanged for this item
		item.dataChanged.connect(
			# lambda *args: print(f"Data changed for item {item._id}")
			lambda item_id=cur_item_id: self.dataChanged.emit(
				# self.index(0, 0),
				# self.index(self.rowCount()-1, self.columnCount()-1))
				self.index(list(self._id_item_map.keys()).index(item_id), 0),
				self.index(list(self._id_item_map.keys()).index(item_id), self.columnCount()-1))
		)



	def append_row(self, item : RunQueueConsoleItem):
		"""Append a row to the model - consisting of a single ConsoleStandardItem
		Args:
			item (ConsoleStandardItem): The item to append (each row corresponds to 1 item)
		"""
		self.beginInsertRows(QtCore.QModelIndex(), self.rowCount(), self.rowCount()) #No parent, insert at end
		self._append_row(item)
		self.endInsertRows()

	def removeRow(self, row: int, parent : QtCore.QModelIndex) -> None:
		self.beginRemoveRows(parent, row, row)
		item_id = list(self._id_item_map.keys())[row]
		with self._id_item_map_mutex:
			del self._id_item_map[item_id]
			self._ignored_ids.add(item_id)

		ignored_list = ", ".join([str(i) for i in list(self._ignored_ids)])
		log.info(f"Removed row with id {item_id} - ignorelist: {ignored_list}")
		self.modelReset.emit() #Why is this needed?

	def parent(self, index : QtCore.QModelIndex) -> QtCore.QModelIndex: #pylint: disable=unused-argument
		return QtCore.QModelIndex() #No parents

	def data(self, index : QtCore.QModelIndex, role : QtCore.Qt.ItemDataRole = QtCore.Qt.ItemDataRole.DisplayRole):
		#Check if index is valid
		if not index.isValid(): #if index is not valid, return None
			return None

		#Get the item from the index
		item = index.internalPointer()
		assert(isinstance(item, RunQueueConsoleItem)), "Invalid item type in RunQueueConsoleModel"

		if role == QtCore.Qt.ItemDataRole.DisplayRole or role == QtCore.Qt.ItemDataRole.EditRole:
			# return item.data(index.column()) #Return the data (str) of the item
			return item.data(role=role, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.ToolTipRole: #When hovering - also display the full name
			return item.data(role=QtCore.Qt.ItemDataRole.DisplayRole, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.DecorationRole:
			return item.data(role=role, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.UserRole + 1:
			return item
		else:
			return None
		# return super().data(index, role)

	def index(self, row : int, column : int, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> QtCore.QModelIndex:
		"""Return the index of the item in the model specified by the given row, column and parent index.

		Args:
			row (int): The row of the item
			column (int): The column of the item
			parent (QtCore.QModelIndex, optional): The parent index. Defaults to QtCore.QModelIndex().

		Returns:
			QtCore.QModelIndex: The index of the item
		"""
		if not parent.isValid(): #If top-level item (should be all items actually)
			id_keys = list(self._id_item_map.keys())
			return self.createIndex(row, column, self._id_item_map[id_keys[row]])
		else: #If item -> no children
			return QtCore.QModelIndex()


if __name__ == "__main__":
	#pylint: disable=ungrouped-imports, protected-access, redefined-outer-name, missing-function-docstring, invalid-name
	from pyside6_utils.widgets.console_widget import ConsoleWidget
	app = QtWidgets.QApplication([])
	model = RunQueueConsoleModel()
	print("Now running a test-instance of RunQueueConsoleModel")

	widget = ConsoleWidget()
	# widget = QtWidgets.QTableView()
	widget.set_model(model)
	widget.show()
	model.append_row(RunQueueConsoleItem(1, "test1"))
	# app.exec()
	model.new_command_line_output(
		item_id=2,
		item_name="test2",
		item_path="",
		item_edit_dt=datetime.datetime.now(),
		item_filepos=0,
		item_msg = "testmsg2"
	)
	samedt = datetime.datetime.now()
	model.new_command_line_output(
		item_id=3,
		item_name="test3",
		item_path="",
		item_edit_dt=samedt,
		item_filepos=0,
		item_msg = "testmsg3"
	)
	model.new_command_line_output(
		item_id=4,
		item_name="test4",
		item_path="",
		item_edit_dt=samedt,
		item_filepos=0,
		item_msg = "testmsg4"
	)
	model._id_item_map[4]._active_state = False
	cur_text = "\n".join([f"This is a test {i}" for i in range(1, 11)]) + '\nThis should now continue...\n'


	def testfunc(model : RunQueueConsoleModel, start_pos):
		i = 10
		cur_pos = start_pos
		while True:
			i+= 1
			new_msg = f"Currently logging {i}\n"
			model.new_command_line_output(
				item_id=1,
				item_name="test1",
				item_path="",
				item_edit_dt=datetime.datetime.now(),
				item_filepos=cur_pos,
				item_msg = new_msg
			)
			cur_pos += len(new_msg)
			if 1 in model._id_item_map:
				print(f"Filled text: {model._id_item_map[1]._filled_filepositions}")
			time.sleep(2)

	print("len curtext: ", len(cur_text))
	thread = threading.Thread(target=lambda: testfunc(model, len(cur_text)))
	thread.start()

	def func(model):
		time.sleep(4)
		model.newCommandLineOutput(
			id=1,
			name="test1",
			path="",
			edit_dt=datetime.datetime.now(),
			filepos=0,
			msg = cur_text
		)
	thread2 = threading.Thread(target=lambda: func(model))
	thread2.start()

	app.exec()



"""An example run function that takes about 20 seconds to complete"""

import logging
import time

from configurun.configuration.configuration import Configuration

log = logging.getLogger(__name__)

def example_target_function(config : Configuration, *args, **kwargs): #pylint: disable=unused-argument
	"""Example run function. The run-function always takes a configuration as the first argument.

	The arguments after that are the arguments passed to the ```RunQueue._process_queue_item()```-method.

	args:
		config (Configuration): The configuration to use
		*args: The arguments passed to the ```RunQueue._process_queue_item()```-method.
		**kwargs: The keyword arguments passed to the ```RunQueue._process_queue_item()```-method.

	returns:
		(None): Nothing is returned
	"""
	log.info("The example run function was called with the following configuration.options classes:")
	for key, value in config.options.items():
		log.info(f"{key} : {type(value).__name__}")

	log.info("We will now be simulating a task that takes about 20 seconds to complete")

	for i in range(20):
		log.info(f"Progress: {i}/20")
		time.sleep(1)

	log.info("Done with the example run function... Now exiting...")



"""Implements a function to convert an argparse.ArgumentParser to a dataclass which can be used as an option source."""



import argparse
import logging
import sys
from dataclasses import field, make_dataclass

from pyside6_utils.classes.constraints import (Interval, Options)

from configurun.configuration.base_options import BaseOptions

log = logging.getLogger(__name__)


def argparse_to_dataclass(
		argument_parser : argparse.ArgumentParser,
		dataclass_name : str = "ArgparseOptions",
	) -> type:
	"""Convert an argparse.ArgumentParser to a dataclass which can be used as an option source.

	NOTE: when this function is called, it automatically adds the dataclass_name to the global scope,
	overwriting any existing class with the same name.

	NOTE: if using a server, make sure to also define this class on the server-side, otherwise the server/client 
	communication will fail.
	
	Args:
		argument_parser (argparse.ArgumentParser): The argparse.ArgumentParser to convert

	Returns:
		type: The dataclass
	"""
	dataclass_args = []
	for action in argument_parser._actions: #pylint: disable=protected-access #NOTE: protected access is neccesary here
			#to access the properties of the actions
		#Iterate over parser args, and create a dataclass field for each
		if action.dest == 'help': #Skip help
			continue

		new_type = action.type
		new_constraints = [new_type]
		if action.choices is not None: #if limited to choices,
			#If range
			if isinstance(action.choices, range):
				new_constraints = [Interval(new_type, action.choices.start, action.choices.stop, closed='both')]
			else:
				new_constraints = [Options(new_type, set(action.choices))] #If list of options
		elif new_type is None: #If no choice but also no type
			if isinstance(action, argparse._StoreTrueAction): #pylint: disable=protected-access
				new_type = bool
				new_constraints = [bool]

		display_name = action.dest.replace('_', ' ').title() #Convert varname to title-like display name
		if action.metavar: #If a specific display name is given, use that instead
			display_name = action.metavar

		if action.nargs is not None and int(action.nargs) > 1:
			raise NotImplementedError("Nargs > 1 not yet implemented for conversion to dataclass")

		dataclass_args.append( #Append tuples of (name, type, field(<neccesary data for UI>))
			(
				action.dest, #Name of the parameter
				action.type,
				field(
					default=action.default,
					metadata=dict(
						display_name = display_name,
						help = action.help,
						constraints = new_constraints, #TODO: add constraints
						required = action.required
					)
				)
			)
		)

	new_dataclass = make_dataclass(
		dataclass_name,
		dataclass_args,
		bases=(BaseOptions,)
	)
	module = None
	try: #NOTE: this should be fixed in Python version 3.12.0 alpha 7
			# (https://github.com/python/cpython/commit/b48be8fa18518583abb21bf6e4f5d7e4b5c9d7b2)
			#make_dataclass now has a module parameter which fixes this issue (even if not specified)
		module = sys._getframemodulename(0) or '__main__' #pylint: disable=protected-access #type:ignore
	except AttributeError:
		try:
			module = sys._getframe(0).f_globals.get('__name__', '__main__') #pylint: disable=protected-access
		except (AttributeError, ValueError):
			log.error(f"Could not set module attribute of dataclass {new_dataclass.__name__}")
	new_dataclass.__module__ = module #Set the module attribute of the new class to the module of the caller #type:ignore
	globals()[dataclass_name] = new_dataclass #Also add class to global scope to enable pickling from this module
		# note that it is probably easier to make use of dill
	return new_dataclass


if __name__ == "__main__":
	from configurun.app.create_run import run_local
	from configurun.examples.example_argparse import parser_example
	test_dataclass = argparse_to_dataclass(parser_example)
	run_local(
		target_function = lambda x, *_: print(x),
		options_source = test_dataclass
	)



"""
Forms a pair with RunQueueClient - this class runs on the server side and handles all incoming connections.
It contains a RunQueue object to/from which all data is transmitted/received using the connected clients.
"""


import hashlib
import logging
import os
import socket
import threading
import time
import typing

import dill
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
import PySignal

from configurun.classes.run_queue import (WORKSPACE_RUN_QUEUE_SAVE_NAME,
                                          RunQueue, WorkspaceInUseException)
from configurun.classes.run_queue_datatypes import (
    AES_EMPTY_KEY, RSA_KEY_SIZE_BITS, AESSessionKeyTransmissionData,
    AuthenticationException, ClientData, LoginTransmissionData,
    PickledDataType, PickleTransmissionData, PubKeyTransmissionData,
    StateMsgType, StateTransmissionData, Transmission, TransmissionType)

log = logging.getLogger(__name__)
DATA_RECEIVE_TIMEOUT = 0.5 #Seconds of checking for new data before timing out to check if the server should stop

class RunQueueServer():
	"""
	A wrapper around a runqueue that opens up a socket an listens for a client to connect.
	RunQueueClient will run client-side and will emulate an instance of the RunQueue class by interfacing with the
	actual runqueue running on the server.

	NOTE: this class assumes that all information for the runqueue is transmitted to the ui via signals - or by
	the ui explicitly calling the functions of the runqueue. All signals in the runqueue are automatically connected
	to the emit_signal function of this class. All function calls on the client side are automatically forwarded to the
	server side.

	Call RunQueueServer.run() after instantiating the class to start the server.
	"""
	_salt = b"\xe6F\xd9\x8f\x15tS5H\xd2e\x82='\x18w\xac\xf1\xfd\x0c\x9c\x89\xf3rMX\xfc\xa1\xea\x8cm\xa2" #Random salt

	def __init__(self,
			run_queue: RunQueue,
			password : str,
			hostname : str = "",
			port : int = 5454,
			workspace_path : str = ""
		) -> None:
		"""
		Args:
			run_queue (RunQueue): The runqueue to use
			password (str): The password to use for authentication
			hostname (str, optional): The hostname to bind to. Defaults to "".
			port (int, optional): The port to bind to. Defaults to 5454.
			workspace_path (str, optional): The path to the workspace.
				Defaults to "". If none/empty, the workspace will be saved/loaded from/to '~/Configurun-Server/'
		"""

		self._run_queue = run_queue #Thread-safe run-queue, as long as we interface using only the provided functions


		self._workspace_path = workspace_path
		if workspace_path is None or len(workspace_path) == 0:
			self._workspace_path = os.path.join(os.path.expanduser("~"), "Configurun-Server")
			log.info(f"Using default workspace path: {self._workspace_path}")

		if not os.path.isdir(self._workspace_path):
			log.info(f"Creating workspace directory {self._workspace_path}")
			os.makedirs(self._workspace_path)

		try:
			self._run_queue.set_workspace_lock(self._workspace_path)
		except WorkspaceInUseException:
			log.warning(f"Workspace {self._workspace_path} is already in use")
			#Ask the user if they want to continue using the workspace using cmd input
			while True:
				choice = input(f"Workspace {self._workspace_path} seems to be in use by another instance of the "
		   			f"Configurun-app. Using this workspace anyway might overwrite existing save-files, or output "
					f" of this process might be overwritten by the other process. Do you want to continue anyway? (y/n): ")
				if choice.lower() == "y":
					break
				elif choice.lower() == "n":
					#re-raise the exception
					raise
				else:
					print("Invalid choice - please enter 'y' or 'n'")

		self._initial_run_queue_load() #Load the run-queue from the workspace- might ask for user input if problem arises

		self._socket = None

		#Generate a public/private key pair for the server
		self._rsa_key : RSA.RsaKey = RSA.generate(RSA_KEY_SIZE_BITS)

		self._public_key = self._rsa_key.publickey().export_key()
		self._private_key = self._rsa_key.export_key()
		self._private_cipher = PKCS1_OAEP.new(self._rsa_key) #Used to decrypt data sent by the clients

		self._client_public_keys = {}
		self._cipher = None

		self.set_password_hash(self.get_password_hash(password=password))

		self._client_dict : dict[socket.socket, ClientData]= {}
		self._client_dict_lock = threading.Lock()

		self._connection_listener_thread = None
		self._is_terminating = False #To check whether the server is terminating (don't start doing new stuff)
		self._server_stop_flag = threading.Event() #To stop the server from running (does not always indicate termination)

		self._hostname = hostname
		self._port = port

		#============= Manage signal-triggered events =============
		for signal_name in self._run_queue.__class__.__dict__:
			# if isinstance(self._run_queue.__class__.__dict__[signal_name], QtCore.Signal): #Previous version used qt
			if isinstance(self._run_queue.__class__.__dict__[signal_name], PySignal.ClassSignal): #Use pysignal for server-side
				target_signal = getattr(self._run_queue, signal_name)
				target_signal.connect(lambda *args, signal_name=signal_name: self.emit_signal(signal_name, *args))

		log.info("Initialized RunQueueServer")

	def terminate(self):
		"""
		Closes the server socket and terminates all clients, saves the current run-queue to the workspace.
		"""
		#First check if there are any items running in the run-queue
		if self._run_queue.get_running_configuration_count() > 0:
			while True:
				user_input = input("There are still items running in the run-queue. Do you want to terminate anyway? (y/n): ")
				if user_input.lower() == "y":
					break
				elif user_input.lower() == "n":
					log.info("Server-termination cancelled.")
					return
				else:
					print("Invalid input - please enter 'y' or 'n'")
		#Save the run-queue to the workspace
		queue_dict = self._run_queue.get_queue_contents_dict(save_running_as_stopped=True)

		dill.dump(queue_dict, open(os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME), "wb"))
		log.info("Terminating server...")
		self._run_queue.stop_autoprocessing() #Indicate to server that it should stop processing new items


		self._is_terminating = True #Make sure we don't start any new threads
		self._server_stop_flag.set() #Stop the connection-listener thread
		try:
			self.disconnect_all_clients()
		except TimeoutError:
			log.error("TimeoutError while waiting for clients to disconnect - continuing with shutdown anyway. "
		 		"Forcefully terminating socket...")

		if self._connection_listener_thread and self._connection_listener_thread.is_alive():
			self._connection_listener_thread.join(3) #Wait for the connection-listener thread to stop
			if self._connection_listener_thread.is_alive():
				log.error("Connection-listener thread did not stop in time... Quitting anyway.")

		if self._socket:
			self._socket.close()
			self._socket = None

		self._run_queue.stop_command_line_queue_emitter() #Stop the command-line queue emitter
		self._run_queue.release_workspace_lock(self._workspace_path)
		log.info("Server terminated")


		# self._run_queue.(os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME))


	def _initial_run_queue_load(self):
		"""
		Should be called just after setting workspace.
		Tries to load the RunQueue contents from a file according to the filename that is used when the close-event is
		called. If the file does not exist, nothing happens.
		"""
		if os.path.exists(os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)):
			cur_items = self._run_queue.get_all_items_dict_snapshot_copy()
			if len(cur_items.keys()) > 0:
				#Ask user if he/she wants to overwrite.
				while True:
					user_input = input("Trying to load run-queue from the workspace, but the current run-queue is not "
					"empty. Do you want to load the run-queue from the workspace anyway? This will overwrite the "
					"currently loaded queue (y/n): ")
					if user_input.lower() == "y":
						break
					elif user_input.lower() == "n":
						raise RuntimeError("Cannot load run-queue from workspace - current run-queue is not empty")
					else:
						print("Invalid input - please enter 'y' or 'n'")

			path = os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)
			queue_contents_dict = dill.load(open(path, "rb"))
			self._run_queue.load_queue_contents_dict(queue_contents_dict)
			log.info(f"Loaded run-queue from {path}")
			# self.ui.runQueueWidget.load_from_file_popup
		else:
			log.info(f"No RunQueue save file found at {os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)}"
					", continuing with the current (empty) run-queue.")

	def stop(self):
		"""Inidicates to all running threads that the server should stop the connection
		NOTE: we can still call run() after calling stop() - this will restart the server
		"""
		self.disconnect_all_clients()
		self._server_stop_flag.set()


	def run(self):
		"""Starts the server and listens for connections
		"""
		if self._is_terminating:
			raise RuntimeError("Server is terminating - cannot start running the server now...")
		if self._connection_listener_thread is not None:
			raise RuntimeError("Server is already running")
		if self._socket is not None:
			raise RuntimeError("Server is already running - socket already exists")
		self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		used_host_name = socket.gethostname() if (self._hostname is None or len(self._hostname) == 0) else self._hostname
		log.info(f"Binding server to {used_host_name}:{self._port}")
		self._socket.bind(( #Bind the socket to the given hostname and port
				used_host_name,
				self._port
		))
		self._connection_listener_thread = threading.Thread(target=self._listen_for_connections)
		self._connection_listener_thread.start()
		self._server_stop_flag.clear()
		# self._listen_for_connections()

	def emit_signal(self, signal_name: str, *args):
		"""Emits a QtCore.Signal to the server

		Args:
			signal_name (str): The signal name
			args*: The arguments to the signal
		"""
		#Print the signal name and args (but max 100 characters)
		log.debug(f"Emitting signal {signal_name} with args: {args if len(str(args)) < 100 else str(args)[:100] + '...'}")

		if self._socket is None:
			log.error("Cannot emit signal - server is not running")
			return

		pickle_data = PickleTransmissionData(
			(PickledDataType.SIGNAL_EMIT, (signal_name, args))
		)
		for client in self._client_dict.values(): #Send all clients the signal
			Transmission.send(
				client.client_socket,
				pickle_data,
				client.client_session_aes_key
			)


	def set_password_hash(self, password_hash : str, invalidate_clients : bool = False):
		"""Sets the password hash to the given value
		if invalidate_clients is True, all clients will be disconnected

		Args:
			password_hash (str): The new password hash
			invalidate_clients (bool, optional): If True, all clients will be disconnected. Defaults to False.
		"""
		if invalidate_clients:
			raise NotImplementedError("Invalidating clients is not yet implemented")
		self._password_hash = password_hash

	@staticmethod
	def get_password_hash(password : str):
		"""Returns the (salted) password hash for the given password

		Args:
			password (str): The password to hash
		"""
		return hashlib.sha512(RunQueueServer._salt + password.encode('utf-8')).hexdigest()


	def _listen_for_connections(self):
		"""To be ran in a separate thread.
		Continually listens for new connections and launches a new thread to authenticate them if found using
		_connection_setup and a new thread.

		#TODO: stopflag/create separate listener class with a qt signal
		"""
		while not self._server_stop_flag.is_set():
			if not self._socket:
				log.warning("Server socket is null - cannot listen for connections - stopping...")
				return
			self._socket.settimeout(DATA_RECEIVE_TIMEOUT) #Set a timeout so we can check if we should stop
			try:
				self._socket.listen(5) #TODO: maybe increase the backlog?
				client, address = self._socket.accept() #TODO: as soon as a client connects, start a new thread to handle
			except (socket.timeout, TimeoutError):
				continue

				#  the new socket as to not block other clients from connecting
			log.info(f"Initial (unauthenticated) connection from {address} has been established - now delegating to \
						a new thread to handle the connection")
			authenticator_thread = threading.Thread(target=self._connection_setup, args=(client, address))
			authenticator_thread.start()



	def _connection_setup(self, client_sock : socket.socket, address : str):
		"""Continuously accept connections and try to initiate a handshake

		The following steps are performed:
		1. Accept a connection from a new client - the client sends its public key (2048 bits)
		2. Return the server public key to the client (encrypted with the client public key) (2048 bits)
		3. Receive the encrypted password from the client & try to decrypt it (512 bits)
		4. If incorrect - stop & send a failure message
		5. If correct, continuously receive data from the client, process it and send the result back

		Args:
			client_sock (socket.socket): The client socket
			address (str): The client address
		"""
		log.info(f"Started authentication process of new client ({address})")
		#Attempt to get client public key
		received = Transmission.receive(client_sock) #First message should be a TransmissionType.pubkey
		client_public_key = None
		if self._is_terminating:
			log.info(f"Server is terminating - could not start connection-setup with {address}")
			return


		if received.transmission_type == TransmissionType.PUB_KEY:
			try:
				pubkey_data = PubKeyTransmissionData.from_transmission_bytes(received.transmission_data)
			except AuthenticationException as auth_exception:
				log.info(f"Error during Authentication of {address}: First transmission from client should be a public\
		 			key, but failed to parse it as such. {auth_exception} - disconnecting...")
				client_sock.close()
				return
			client_public_key = pubkey_data.pubkey
		elif received.transmission_type == TransmissionType.STATE:
			try:
				state_data = StateTransmissionData.from_transmission_bytes(received.transmission_data)
				log.warning(f"Authentication of {address}: Received a state message from {address}: \
					{state_data.state_msg_type.name} - {state_data.state_msg}... But first message should be a public \
					key. Closing connection...")
				client_sock.close()
			except Exception as exception: # pylint: disable=broad-exception-caught
				log.warning(f"First connection should be a public key - failed to parse {address}: {exception} - \
					disconnecting...")
				client_sock.close()
				return
		else:
			log.warning(f"Error during Authentication of {address}: Received a message with TransmissionType: \
		   		{received.transmission_type} - expected TransmissionType.pub_key - disconnecting...")

		if client_public_key is None:
			log.error(f"Error during Authentication of {address}: No public key received in pubkey transmission - \
		 		disconnecting...")
			Transmission.send(
				client_sock,
				StateTransmissionData(
					StateMsgType.ERROR,
					"Failure - no public key received from client - first response message must be a public key"
				)
			)
			return
		log.info(f"Received public key from client {address}: {client_public_key}")

		#Instantiate a rsa key object from the client public key
		client_public_key = RSA.import_key(client_public_key)

		rsa_cipher = PKCS1_OAEP.new(client_public_key)

		#Generate a random AES-session key
		aes_session_key = get_random_bytes(32)

		#Encrypt the AES key with the client public key, and send it to the client
		Transmission.send(
			client_sock,
			AESSessionKeyTransmissionData(
				encrypted_session_key=rsa_cipher.encrypt(aes_session_key) #Encrypt
			)
		)

		log.info(f"Authentication of {address}: session key is {aes_session_key}")

		# ====== FROM NOW ON - ONLY ENCRYPTED MESSAGES SHOULD BE SENT (USING AES) ===== (both parties have the AES key)
		received = Transmission.receive(
			client_sock,
			aes_session_key
		)
		if received.transmission_type != TransmissionType.LOGIN:
			log.error(f"Error during Authentication of {address}: Expected TransmissionType.login - received \
		 		{received.transmission_type.name}({received.transmission_type}) - disconnecting...")
			client_sock.close()
			return
		else:
			log.info(f"Authentication of {address}: Received TransmissionType.login from {address}")
			login_info = LoginTransmissionData.from_transmission_bytes(received.transmission_data)

		hashed_password = self.get_password_hash(login_info.password)
		log.info(f"Authentication of {address} - Received hashed pw {address}: {hashed_password}")
			#TODO: Also implement client-side hashing of the password - send a salt back with the AES-session key?
			# Traffic is encrypted - but server probably shouldn't know the plaintext password attempts.


		#Check if the password hash is correct
		if hashed_password != self._password_hash:
			log.error(f"Error during Authentication of {address}: Invalid password received from {address} - disconnecting...")
			Transmission.send(
				client_sock,
				StateTransmissionData(
					StateMsgType.LOGIN_ERROR,
					"Failure - invalid password hash received from client"
				),
				aes_session_key
			)
			client_sock.close()
			return
		else:
			Transmission.send(
				client_sock,
				StateTransmissionData(
					StateMsgType.LOGIN_ACCEPTED,
					"Success - client authenticated"
				),
				aes_session_key
			)

		ip_list = [client.client_socket.getpeername() for client in self.list_authenticated_clients()]
		log.info(f"Connection with {address} has now been authenticated - starting a new thread to handle the \
	   		client/server interaction. Currently connected clients ({len(ip_list)}): \
	   		{ip_list }")

		#Start a new thread to handle the client/server interaction #TODO: create worker with a qt signal
		client_stop_flag = threading.Event()
		client_stop_flag.clear()
		client_thread = threading.Thread(
			target=self._client_listener,
			args=(
				client_sock,
				address,
				aes_session_key,
				client_stop_flag
			)
		)

		with self._client_dict_lock:
			assert client_sock not in self._client_dict, "Connection cannot already be authenticated/running"
			self._client_dict[client_sock] = ClientData(
				client_socket = client_sock,
				client_public_key = client_public_key,
				client_session_aes_key = aes_session_key,
				client_listening_thread = client_thread,
				client_listening_thread_stop_flag=client_stop_flag
			)
		client_thread.start()

		return #Should end the authentication thread


	def list_authenticated_clients(self) -> typing.List[ClientData]:
		"""Returns a list of all currently connected authenticated clients
		"""
		with self._client_dict_lock:
			return list(self._client_dict.values())


	def _handle_pickled_data(self,
			client : socket.socket,
			client_address : str,
			aes_session_key : bytes,
			pickle_transmission_data: PickleTransmissionData):
		"""
		Handles a pickled object received from a client
		"""
		unpickled_data = pickle_transmission_data.unpickled_data

		assert isinstance(unpickled_data, tuple), "Pickled transmission data should (in this case) always be a tuple of \
			the form (pickledDataType, data)"
		# assert(unpickled_data[0] == pickledDataType.function_call)

		error_msg = None
		if unpickled_data[0] == PickledDataType.METHOD_CALL:
			function_data = unpickled_data[1]
			#Handle a function call
			method_call_id = function_data[0] #So client can identify which response belongs to which call
			function_name = function_data[1]
			args = function_data[2]
			kwargs = function_data[3]
			log.debug(f"Received a function call from a client: {function_name}(self, {args}, {kwargs})")
			try:
				result = getattr(self._run_queue, function_name)(*args, **kwargs)
			except Exception as exception: # pylint: disable=broad-exception-caught
				result = exception
				log.warning(f"Client {client_address} called function {function_name}, which result in \
		 			{type(exception)}: {exception}")

			#Send the result back to the client
			Transmission.send(
				client,
				PickleTransmissionData(
					(PickledDataType.METHOD_RETURN, (method_call_id, result)) #return the call_id and the results
				),
				aes_cypher_key=aes_session_key,
			)
		elif unpickled_data[0] == PickledDataType.SIGNAL_EMIT:
			error_msg = "Received a signal emit from client - this should not be possible - ignoring transmission"
		elif unpickled_data[0] == PickledDataType.METHOD_RETURN:
			error_msg = "Received a function return from client - this should not be possible - ignoring transmission"
		else:
			error_msg = f"Received a pickled object with unknown type: {unpickled_data[0]} - ignoring transmission"

		if error_msg is not None: #Log/send and raise and error if necessary
			Transmission.send(
				client,
				StateTransmissionData(
					StateMsgType.ERROR,
					error_msg
				),
				aes_cypher_key=aes_session_key,
			)
			raise TypeError(error_msg)

	def disconnect_all_clients(self, timeout_seconds : float = 5.0):
		"""
		Attempts to disconnect all clients by sending them a disconnect signal and waiting for them to disconnect.
		Will wait for timeout_seconds seconds for each client to disconnect before giving up and raising a TimeoutError.
		"""
		start_time = time.time()
		with self._client_dict_lock:
			for client_data in self._client_dict.values(): #Flag all clients to stop
				client_data.client_listening_thread_stop_flag.set()

			for client_data in self._client_dict.values(): #Then wait for them to stop
				remaining_time = timeout_seconds - (time.time() - start_time)
				if client_data.client_listening_thread.is_alive(): #Wait for all living client threads to stop
					client_data.client_listening_thread.join(timeout=remaining_time)

				if client_data.client_listening_thread.is_alive(): #If the thread is still alive, it didn't stop in time
					raise TimeoutError(f"Client {client_data.client_socket.getpeername()} did not stop in time...")

			self._client_dict = {}



	def _client_listener(self,
				client : socket.socket,
				client_address : str,
				aes_session_key : bytes,
				client_stop_flag : threading.Event
			):
		"""
		NOTE: this function assumes that the client has already been authenticated and should only be called when this
		is the case. Listens to the given client and handles all incoming requests

		Args:
			client (socket.socket): The client to listen to
			aes_session_key (bytes): The AES session key used to encrypt data between the client and the server
				(required - all data should be authenticated)

		"""
		assert (aes_session_key is not None) and (aes_session_key != AES_EMPTY_KEY), "Session_key can't be null/empty \
			when listening to a client"
		try:
			while not client_stop_flag.isSet(): #TODO: put in a worker so it shuts down automtaically on server close
				try:
					#Receive the transmission data - this should always be succesful as long as the connection is active
					received = Transmission.receive(client, aes_session_key, timeout_seconds=DATA_RECEIVE_TIMEOUT)
					log.info(f"Received transmission from client {client_address} with TransmissionType: \
						{received.transmission_type.name}({received.transmission_type})")

					if received.transmission_type == TransmissionType.PICKLED_OBJECT:
						try:
							pickled_data = PickleTransmissionData.from_transmission_bytes(received.transmission_data)
						except Exception as exception: # pylint: disable=broad-exception-caught
							log.error(f"Failed to unpickle data from client {client_address}: {exception}")
							Transmission.send(
								client,
								StateTransmissionData(
									StateMsgType.ERROR,
									f"Failed to unpickle data from client {client_address}: {exception}"
								)
							)
							continue
						try:
							self._handle_pickled_data(client=client,
								client_address=client_address,
								aes_session_key=aes_session_key,
								pickle_transmission_data=pickled_data
							)
						except Exception as exception: # pylint: disable=broad-exception-caught
							log.error(f"Failed to handle pickled data from client {client_address}: {exception} ")
					elif received.transmission_type == TransmissionType.STATE:
						data = StateTransmissionData.from_transmission_bytes(received.transmission_data)
						log.info(f"State Client : {data.state_msg_type.name} - {data.state_msg}")
					else:
						log.error(f"Received a transmission with TransmissionType: {received.transmission_type.name}\
							({received.transmission_type}) - this should not happen for authenticated clients \
							- ignoring transmission.")
				except TimeoutError:
					pass #Continue
		except Exception as exception: # pylint: disable=broad-exception-caught
			log.error(f"Error during client/server interaction: {exception} - disconnecting client {client_address}")
			with self._client_dict_lock: #Remove the client from the list of authenticated clients
				del self._client_dict[client]
			client.close() #Try to close connection (if not already closed)


def run_example_server(password : str = "password", port : int = 5454, log_level : int = logging.INFO):
	"""
	Runs an example-instance of the server, using the example_target_function from 
	configurun/examples/example_target_function.py and the passed password as well as the default
	workspace path (~/Configurun-Server)

	Args:
		password (str, optional): The password to connect to the server. Defaults to "password".
	"""
	#pylint: disable=import-outside-toplevel
	from configurun.server.create_server import run_server
	import tempfile
	from configurun.examples.example_target_function import example_target_function
	tempdir = tempfile.gettempdir()
	workspace_path = os.path.join(tempdir, "Configurun", "Configurun-Server-Example")

	run_server(
		target_function=example_target_function,
		workspace_path=workspace_path,
		password=password,
		port=port,
		log_level=log_level
	)


if __name__ == "__main__":
	run_example_server(log_level=logging.DEBUG)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.5.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QDockWidget, QGridLayout, QHBoxLayout,
    QHeaderView, QMainWindow, QMenu, QMenuBar,
    QPushButton, QSizePolicy, QStatusBar, QToolButton,
    QUndoView, QVBoxLayout, QWidget)

from configurun.app.widgets.run_queue_widget import RunQueueWidget
from pyside6_utils.widgets.console_widget import ConsoleWidget
from pyside6_utils.widgets.extended_mdi_area import ExtendedMdiArea
from pyside6_utils.widgets.file_explorer_view import FileExplorerView
from pyside6_utils.widgets.overlay_widget import OverlayWidget
from pyside6_utils.widgets.square_frame import SquareFrame
import configurun.res.app_resources_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(2112, 1141)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/apps/utilities-system-monitor.png", QSize(), QIcon.Normal, QIcon.Off)
        MainWindow.setWindowIcon(icon)
        self.actionUndo = QAction(MainWindow)
        self.actionUndo.setObjectName(u"actionUndo")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/actions/edit-undo.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionUndo.setIcon(icon1)
        self.actionRedo = QAction(MainWindow)
        self.actionRedo.setObjectName(u"actionRedo")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/actions/edit-redo.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionRedo.setIcon(icon2)
        self.actionIncreaseFontSize = QAction(MainWindow)
        self.actionIncreaseFontSize.setObjectName(u"actionIncreaseFontSize")
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/actions/list-add.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionIncreaseFontSize.setIcon(icon3)
        self.actionDefaultFontSize = QAction(MainWindow)
        self.actionDefaultFontSize.setObjectName(u"actionDefaultFontSize")
        icon4 = QIcon()
        icon4.addFile(u":/Icons/icons/actions/view-refresh.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionDefaultFontSize.setIcon(icon4)
        self.actionDecreaseFontSize = QAction(MainWindow)
        self.actionDecreaseFontSize.setObjectName(u"actionDecreaseFontSize")
        icon5 = QIcon()
        icon5.addFile(u":/Icons/icons/actions/list-remove.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionDecreaseFontSize.setIcon(icon5)
        self.actionSave = QAction(MainWindow)
        self.actionSave.setObjectName(u"actionSave")
        icon6 = QIcon()
        icon6.addFile(u":/Icons/icons/actions/document-save.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSave.setIcon(icon6)
        self.actionSave_As = QAction(MainWindow)
        self.actionSave_As.setObjectName(u"actionSave_As")
        icon7 = QIcon()
        icon7.addFile(u":/Icons/icons/actions/document-save-as.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSave_As.setIcon(icon7)
        self.actionReset_Splitters = QAction(MainWindow)
        self.actionReset_Splitters.setObjectName(u"actionReset_Splitters")
        icon8 = QIcon()
        icon8.addFile(u":/Icons/icons/Tango Icons/32x32/actions/view-refresh.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionReset_Splitters.setIcon(icon8)
        self.actionSetLocalRunMode = QAction(MainWindow)
        self.actionSetLocalRunMode.setObjectName(u"actionSetLocalRunMode")
        self.actionSetLocalRunMode.setCheckable(True)
        self.actionSetLocalRunMode.setChecked(True)
        self.actionSetLocalRunMode.setEnabled(True)
        icon9 = QIcon()
        icon9.addFile(u":/Icons/icons/Tango Icons/32x32/actions/go-home.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSetLocalRunMode.setIcon(icon9)
        self.actionSetNetworkRunMode = QAction(MainWindow)
        self.actionSetNetworkRunMode.setObjectName(u"actionSetNetworkRunMode")
        self.actionSetNetworkRunMode.setCheckable(True)
        icon10 = QIcon()
        icon10.addFile(u":/Icons/icons/Tango Icons/32x32/apps/internet-web-browser.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSetNetworkRunMode.setIcon(icon10)
        self.actionNewConfig = QAction(MainWindow)
        self.actionNewConfig.setObjectName(u"actionNewConfig")
        icon11 = QIcon()
        icon11.addFile(u":/Icons/icons/actions/document-new.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionNewConfig.setIcon(icon11)
        self.actionNone = QAction(MainWindow)
        self.actionNone.setObjectName(u"actionNone")
        self.actionNone.setEnabled(False)
        self.actionReset_Splitters_2 = QAction(MainWindow)
        self.actionReset_Splitters_2.setObjectName(u"actionReset_Splitters_2")
        icon12 = QIcon()
        icon12.addFile(u":/Icons/icons/mimetypes/x-office-document-template.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionReset_Splitters_2.setIcon(icon12)
        self.actionBackupRunQueue = QAction(MainWindow)
        self.actionBackupRunQueue.setObjectName(u"actionBackupRunQueue")
        self.actionLoadRunQueue = QAction(MainWindow)
        self.actionLoadRunQueue.setObjectName(u"actionLoadRunQueue")
        self.action_None = QAction(MainWindow)
        self.action_None.setObjectName(u"action_None")
        self.actionOpenConfig = QAction(MainWindow)
        self.actionOpenConfig.setObjectName(u"actionOpenConfig")
        icon13 = QIcon()
        icon13.addFile(u":/Icons/icons/actions/document-open.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionOpenConfig.setIcon(icon13)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(9, 0, 0, 0)
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, -1, -1, -1)
        self.ConfigurationMdiArea = ExtendedMdiArea(self.centralwidget)
        self.ConfigurationMdiArea.setObjectName(u"ConfigurationMdiArea")

        self.verticalLayout_2.addWidget(self.ConfigurationMdiArea)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.addToQueueButton = QPushButton(self.centralwidget)
        self.addToQueueButton.setObjectName(u"addToQueueButton")
        icon14 = QIcon()
        icon14.addFile(u":/Icons/icons/actions/format-indent-more.png", QSize(), QIcon.Normal, QIcon.Off)
        self.addToQueueButton.setIcon(icon14)

        self.horizontalLayout_2.addWidget(self.addToQueueButton)

        self.saveToQueueItemBtn = QPushButton(self.centralwidget)
        self.saveToQueueItemBtn.setObjectName(u"saveToQueueItemBtn")
        icon15 = QIcon()
        icon15.addFile(u":/Icons/icons/actions/savesymbol.png", QSize(), QIcon.Normal, QIcon.Off)
        self.saveToQueueItemBtn.setIcon(icon15)

        self.horizontalLayout_2.addWidget(self.saveToQueueItemBtn)


        self.verticalLayout_2.addLayout(self.horizontalLayout_2)


        self.verticalLayout.addLayout(self.verticalLayout_2)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 2112, 22))
        self.menuasdf = QMenu(self.menubar)
        self.menuasdf.setObjectName(u"menuasdf")
        self.menuview = QMenu(self.menubar)
        self.menuview.setObjectName(u"menuview")
        self.menuSet_Font_Size = QMenu(self.menuview)
        self.menuSet_Font_Size.setObjectName(u"menuSet_Font_Size")
        self.menuSet_Font_Size.setGeometry(QRect(0, 0, 144, 122))
        icon16 = QIcon()
        icon16.addFile(u":/Icons/icons/apps/preferences-desktop-font.png", QSize(), QIcon.Normal, QIcon.Off)
        self.menuSet_Font_Size.setIcon(icon16)
        self.menuMDI_Area = QMenu(self.menuview)
        self.menuMDI_Area.setObjectName(u"menuMDI_Area")
        self.menuMDI_Area.setEnabled(True)
        icon17 = QIcon()
        icon17.addFile(u":/Icons/icons/apps/preferences-system-windows.png", QSize(), QIcon.Normal, QIcon.Off)
        self.menuMDI_Area.setIcon(icon17)
        self.menuRun_Queue = QMenu(self.menubar)
        self.menuRun_Queue.setObjectName(u"menuRun_Queue")
        self.actionViewRunQueueFilter = QMenu(self.menuRun_Queue)
        self.actionViewRunQueueFilter.setObjectName(u"actionViewRunQueueFilter")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.dockWidget = QDockWidget(MainWindow)
        self.dockWidget.setObjectName(u"dockWidget")
        icon18 = QIcon()
        icon18.addFile(u":/Icons/icons/savesymbol.png", QSize(), QIcon.Normal, QIcon.Off)
        self.dockWidget.setWindowIcon(icon18)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName(u"dockWidgetContents")
        self.gridLayout = QGridLayout(self.dockWidgetContents)
        self.gridLayout.setObjectName(u"gridLayout")
        self.verticalLayout_5 = QVBoxLayout()
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.saveCurrentConfigBtn = QPushButton(self.dockWidgetContents)
        self.saveCurrentConfigBtn.setObjectName(u"saveCurrentConfigBtn")
        self.saveCurrentConfigBtn.setIcon(icon6)

        self.horizontalLayout_6.addWidget(self.saveCurrentConfigBtn)

        self.saveCurrentConfigAsBtn = QPushButton(self.dockWidgetContents)
        self.saveCurrentConfigAsBtn.setObjectName(u"saveCurrentConfigAsBtn")
        self.saveCurrentConfigAsBtn.setIcon(icon7)

        self.horizontalLayout_6.addWidget(self.saveCurrentConfigAsBtn)

        self.squareFrame = SquareFrame(self.dockWidgetContents)
        self.squareFrame.setObjectName(u"squareFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.squareFrame.sizePolicy().hasHeightForWidth())
        self.squareFrame.setSizePolicy(sizePolicy)
        self.verticalLayout_13 = QVBoxLayout(self.squareFrame)
        self.verticalLayout_13.setSpacing(0)
        self.verticalLayout_13.setObjectName(u"verticalLayout_13")
        self.verticalLayout_13.setContentsMargins(0, 0, 0, 0)
        self.OpenFileLocationBtn = QToolButton(self.squareFrame)
        self.OpenFileLocationBtn.setObjectName(u"OpenFileLocationBtn")
        sizePolicy.setHeightForWidth(self.OpenFileLocationBtn.sizePolicy().hasHeightForWidth())
        self.OpenFileLocationBtn.setSizePolicy(sizePolicy)
        self.OpenFileLocationBtn.setMinimumSize(QSize(24, 24))
        icon19 = QIcon()
        icon19.addFile(u":/Icons/icons/actions/folder-new.png", QSize(), QIcon.Normal, QIcon.Off)
        self.OpenFileLocationBtn.setIcon(icon19)

        self.verticalLayout_13.addWidget(self.OpenFileLocationBtn)


        self.horizontalLayout_6.addWidget(self.squareFrame)

        self.horizontalLayout_6.setStretch(0, 100)
        self.horizontalLayout_6.setStretch(1, 100)

        self.verticalLayout_5.addLayout(self.horizontalLayout_6)

        self.ConfigFilePickerView = FileExplorerView(self.dockWidgetContents)
        self.ConfigFilePickerView.setObjectName(u"ConfigFilePickerView")

        self.verticalLayout_5.addWidget(self.ConfigFilePickerView)


        self.gridLayout.addLayout(self.verticalLayout_5, 0, 0, 1, 1)

        self.dockWidget.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.dockWidget)
        self.dockWidget_3 = QDockWidget(MainWindow)
        self.dockWidget_3.setObjectName(u"dockWidget_3")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.dockWidget_3.sizePolicy().hasHeightForWidth())
        self.dockWidget_3.setSizePolicy(sizePolicy1)
        icon20 = QIcon()
        icon20.addFile(u":/Icons/icons/Tango Icons/32x32/categories/applications-development.png", QSize(), QIcon.Normal, QIcon.Off)
        self.dockWidget_3.setWindowIcon(icon20)
        self.dockWidgetContents_3 = QWidget()
        self.dockWidgetContents_3.setObjectName(u"dockWidgetContents_3")
        self.verticalLayout_4 = QVBoxLayout(self.dockWidgetContents_3)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.runQueueOverlayWidget = OverlayWidget(self.dockWidgetContents_3)
        self.runQueueOverlayWidget.setObjectName(u"runQueueOverlayWidget")
        sizePolicy.setHeightForWidth(self.runQueueOverlayWidget.sizePolicy().hasHeightForWidth())
        self.runQueueOverlayWidget.setSizePolicy(sizePolicy)
        self.runQueueOverlayWidget.setProperty("overlayHidden", True)
        self.runQueueWidget = RunQueueWidget(self.runQueueOverlayWidget)
        self.runQueueWidget.setObjectName(u"runQueueWidget")
        self.runQueueWidget.setGeometry(QRect(1, 1, 18, 18))
        sizePolicy.setHeightForWidth(self.runQueueWidget.sizePolicy().hasHeightForWidth())
        self.runQueueWidget.setSizePolicy(sizePolicy)
        self.verticalLayout_6 = QVBoxLayout(self.runQueueWidget)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")

        self.verticalLayout_4.addWidget(self.runQueueOverlayWidget)

        self.dockWidget_3.setWidget(self.dockWidgetContents_3)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.dockWidget_3)
        self.UndoStack = QDockWidget(MainWindow)
        self.UndoStack.setObjectName(u"UndoStack")
        self.UndoStack.setWindowIcon(icon1)
        self.dockWidgetContents_4 = QWidget()
        self.dockWidgetContents_4.setObjectName(u"dockWidgetContents_4")
        self.gridLayout_2 = QGridLayout(self.dockWidgetContents_4)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.undoView = QUndoView(self.dockWidgetContents_4)
        self.undoView.setObjectName(u"undoView")

        self.gridLayout_2.addWidget(self.undoView, 0, 0, 1, 1)

        self.UndoStack.setWidget(self.dockWidgetContents_4)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.UndoStack)
        self.ConsoleDockWidget = QDockWidget(MainWindow)
        self.ConsoleDockWidget.setObjectName(u"ConsoleDockWidget")
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName(u"dockWidgetContents_2")
        self.horizontalLayout = QHBoxLayout(self.dockWidgetContents_2)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.ConsoleOverlayWidget = OverlayWidget(self.dockWidgetContents_2)
        self.ConsoleOverlayWidget.setObjectName(u"ConsoleOverlayWidget")
        self.ConsoleOverlayWidget.setProperty("overlayHidden", True)
        self.verticalLayout_15 = QVBoxLayout(self.ConsoleOverlayWidget)
        self.verticalLayout_15.setObjectName(u"verticalLayout_15")
        self.verticalLayout_15.setContentsMargins(-1, 0, -1, 0)
        self.verticalLayout_14 = QVBoxLayout()
        self.verticalLayout_14.setObjectName(u"verticalLayout_14")
        self.consoleWidget = ConsoleWidget(self.ConsoleOverlayWidget)
        self.consoleWidget.setObjectName(u"consoleWidget")
        self.consoleWidget.setProperty("ConsoleWidthPercentage", 88)

        self.verticalLayout_14.addWidget(self.consoleWidget)


        self.verticalLayout_15.addLayout(self.verticalLayout_14)


        self.horizontalLayout.addWidget(self.ConsoleOverlayWidget)

        self.ConsoleDockWidget.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(Qt.BottomDockWidgetArea, self.ConsoleDockWidget)

        self.menubar.addAction(self.menuasdf.menuAction())
        self.menubar.addAction(self.menuRun_Queue.menuAction())
        self.menubar.addAction(self.menuview.menuAction())
        self.menuasdf.addAction(self.actionNewConfig)
        self.menuasdf.addAction(self.actionOpenConfig)
        self.menuasdf.addSeparator()
        self.menuasdf.addAction(self.actionSave)
        self.menuasdf.addAction(self.actionSave_As)
        self.menuasdf.addSeparator()
        self.menuasdf.addAction(self.actionUndo)
        self.menuasdf.addAction(self.actionRedo)
        self.menuview.addAction(self.menuSet_Font_Size.menuAction())
        self.menuview.addAction(self.menuMDI_Area.menuAction())
        self.menuSet_Font_Size.addAction(self.actionIncreaseFontSize)
        self.menuSet_Font_Size.addAction(self.actionDefaultFontSize)
        self.menuSet_Font_Size.addAction(self.actionDecreaseFontSize)
        self.menuMDI_Area.addAction(self.actionNone)
        self.menuRun_Queue.addAction(self.actionViewRunQueueFilter.menuAction())
        self.menuRun_Queue.addSeparator()
        self.actionViewRunQueueFilter.addAction(self.action_None)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Configurun[*]", None))
        self.actionUndo.setText(QCoreApplication.translate("MainWindow", u"Undo", None))
#if QT_CONFIG(tooltip)
        self.actionUndo.setToolTip(QCoreApplication.translate("MainWindow", u"Undo the last edit to the settings", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionUndo.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Z", None))
#endif // QT_CONFIG(shortcut)
        self.actionRedo.setText(QCoreApplication.translate("MainWindow", u"Redo", None))
#if QT_CONFIG(shortcut)
        self.actionRedo.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Y", None))
#endif // QT_CONFIG(shortcut)
        self.actionIncreaseFontSize.setText(QCoreApplication.translate("MainWindow", u"Increase Size", None))
#if QT_CONFIG(shortcut)
        self.actionIncreaseFontSize.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+=", None))
#endif // QT_CONFIG(shortcut)
        self.actionDefaultFontSize.setText(QCoreApplication.translate("MainWindow", u"Default Size", None))
        self.actionDecreaseFontSize.setText(QCoreApplication.translate("MainWindow", u"Decrease Size", None))
#if QT_CONFIG(shortcut)
        self.actionDecreaseFontSize.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+-", None))
#endif // QT_CONFIG(shortcut)
        self.actionSave.setText(QCoreApplication.translate("MainWindow", u"Save...", None))
#if QT_CONFIG(shortcut)
        self.actionSave.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+S", None))
#endif // QT_CONFIG(shortcut)
        self.actionSave_As.setText(QCoreApplication.translate("MainWindow", u"Save As...", None))
#if QT_CONFIG(shortcut)
        self.actionSave_As.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Shift+S", None))
#endif // QT_CONFIG(shortcut)
        self.actionReset_Splitters.setText(QCoreApplication.translate("MainWindow", u"Reset Splitters", None))
        self.actionSetLocalRunMode.setText(QCoreApplication.translate("MainWindow", u"Local", None))
        self.actionSetNetworkRunMode.setText(QCoreApplication.translate("MainWindow", u"Network", None))
        self.actionNewConfig.setText(QCoreApplication.translate("MainWindow", u"New Config...", None))
#if QT_CONFIG(shortcut)
        self.actionNewConfig.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+N", None))
#endif // QT_CONFIG(shortcut)
        self.actionNone.setText(QCoreApplication.translate("MainWindow", u"None", None))
        self.actionReset_Splitters_2.setText(QCoreApplication.translate("MainWindow", u"Reset Splitters", None))
        self.actionBackupRunQueue.setText(QCoreApplication.translate("MainWindow", u"Backup...", None))
        self.actionLoadRunQueue.setText(QCoreApplication.translate("MainWindow", u"Load...", None))
        self.action_None.setText(QCoreApplication.translate("MainWindow", u"(None)", None))
        self.actionOpenConfig.setText(QCoreApplication.translate("MainWindow", u"Open...", None))
        self.addToQueueButton.setText(QCoreApplication.translate("MainWindow", u"Append to Queue", None))
        self.saveToQueueItemBtn.setText(QCoreApplication.translate("MainWindow", u"Save to Queue-Item", None))
        self.menuasdf.setTitle(QCoreApplication.translate("MainWindow", u"Configuration", None))
        self.menuview.setTitle(QCoreApplication.translate("MainWindow", u"View", None))
        self.menuSet_Font_Size.setTitle(QCoreApplication.translate("MainWindow", u"Font Size", None))
        self.menuMDI_Area.setTitle(QCoreApplication.translate("MainWindow", u"MDI Area", None))
        self.menuRun_Queue.setTitle(QCoreApplication.translate("MainWindow", u"Run Queue", None))
        self.actionViewRunQueueFilter.setTitle(QCoreApplication.translate("MainWindow", u"(built during runtime)", None))
        self.dockWidget.setWindowTitle(QCoreApplication.translate("MainWindow", u"File Overview", None))
        self.saveCurrentConfigBtn.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.saveCurrentConfigAsBtn.setText(QCoreApplication.translate("MainWindow", u"Save As...", None))
        self.OpenFileLocationBtn.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.dockWidget_3.setWindowTitle(QCoreApplication.translate("MainWindow", u"Run Queue", None))
        self.UndoStack.setWindowTitle(QCoreApplication.translate("MainWindow", u"Undo Stack", None))
        self.ConsoleDockWidget.setWindowTitle(QCoreApplication.translate("MainWindow", u"Command-Line Output", None))
    # retranslateUi




"""Implements RunQueueClient which forms a pair with RunQueueServer.py - this class acts as if it's an actual RunQueue
instance, but instead sends all function calls to a server on which the actual RunQueue instance is running
"""

import logging
import queue
import socket
import threading
import traceback

import PySignal
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA

from configurun.classes.method_call_interceptor import (
    MethodCallInterceptedMeta, get_class_implemented_methods)
from configurun.classes.run_queue import RunQueue
from configurun.classes.run_queue_datatypes import (
    RSA_KEY_SIZE_BITS, AESSessionKeyTransmissionData, AuthenticationException,
    LoginTransmissionData, PickledDataType, PickleTransmissionData,
    PubKeyTransmissionData, StateMsgType, StateTransmissionData, Transmission,
    TransmissionType)

log = logging.getLogger(__name__)


def get_pysignal_names(the_object : type) -> list[PySignal.ClassSignal]:
	"""
	Get all PySignal.ClassSignal instances of an arbitrary object (by name)	
	"""
	signals = []
	for signal_name in the_object.__dict__:
		if isinstance(the_object.__dict__[signal_name], PySignal.ClassSignal):
			signals.append(signal_name)
	return signals



def no_function(*args, **kwargs): #pylint: disable=unused-argument
	"""Dummy function that does nothing"""
	pass #pylint: disable=unnecessary-pass

class NoAuthenticatedConnectionException(Exception):
	"""Exception raised when a function is called on a RunQueueClient that is not connected to a server"""

class RunQueueClient(
		RunQueue,
		metaclass=MethodCallInterceptedMeta,
		intercept_list=get_class_implemented_methods(RunQueue),
		skip_intercept_list=get_pysignal_names(RunQueue)
	):
	"""
	Each clients acts as if it is a runQueue - connecting to a server on which the actual runQueue is running.
	Each function of the runQueue is intercepted by this class and sent to the server (except the signals).

	NOTE: this class uses ```RunQueue```- which uses pyqt signals - make sure a main event loop is running
	(e.g. app.exec()) when using instances of this class

	NOTE: if the client is not connected to a server - the method-interceptor instead returns None (and logs a warning)

	As soon as a connection is established, the client will start listening to the server for signals and method calls.
	Also, the reset-signal is emitted from the queue, indicating that all models should be reset.
	"""

	authenConnectionStateChanged = PySignal.ClassSignal() #Bool emitted when the authenticated connection state changes -
		# True if connected and authenticated, False if either disconnected or not yet authenticated


	def __init__(
			self
		) -> None:
		"""Initializes the RunQueueClient

		Has no args, since RunQueue will be running on server-side, settings attributes here would be useless.
		All function calls will be intercepted and sent to the server.
		"""
		super().__init__(target_function=no_function)
		self._server_ip = None
		self._server_port = None
		self._socket : socket.socket | None = None
		self._authenticated = False
		self._password = None

		#Generate a public/private key pair for the server
		self._rsa_key : RSA.RsaKey = RSA.generate(RSA_KEY_SIZE_BITS)
		self._public_key = self._rsa_key.publickey()
		self._private_key = self._rsa_key
		self._private_rsa_cipher = PKCS1_OAEP.new(self._rsa_key)

		#Created based on the session key received from the server
		self._listen_thread = None
		self._aes_session_key = None

		# self._method_return_queue = queue.Queue() #When the server responds - the listener thread will put the
			# result in this queue - should
		self._method_response_dict : dict[int, queue.Queue] = {} #When the server responds - the listener thread will
			#put the result in this queue - keys are method ids, values are the result of the method call
		self._method_reponse_dict_lock = threading.Lock()
		self._method_id_counter = 0 #Used to keep track of which function call is which

		self._connected_or_connecting = False #Used to make sure that we're only attempting 1 connection at a time

		self._disconnect_flag = True #Whether to stop listening to the server

		self._await_response_lock = threading.Lock() 


	def is_connected(self):
		"""Returns whether the client is currently trying to connect to the server or is connected to the server

		Returns:
			bool: Whether the client is currently trying to connect to the server or is connected to the server
				This doesn't indicate that the connection has been accepted by the server yet
		"""
		return self._connected_or_connecting

	def is_connected_and_authenticated(self) -> bool:
		"""Returns whether an authenticated connection to the server has been established

		Returns:
			bool: Whether an authenticated connection to the server is currently active
		"""
		return self._authenticated

	def _interceptor(self, function_name: str, *args, **kwargs):

		log.debug(f"Intercepted call to RunQueue function {function_name}")# with args: {args} and kwargs: {kwargs}")
		if self._socket is None or not self._authenticated or self._aes_session_key is None:
			# raise NoAuthenticatedConnectionException(f"Could not pass on method call to {function_name}, to server, "
			# 	"as no (authenticated) connection to server has been established yet - returning None")
			return None
		# Send a state message to the server of the intercepted call
		# Transmission.send(
		# 	self._socket,
		# 	StateTransmissionData(
		# 		state_msg_type = stateMsgType.general_msg,
		# 		state_msg = f"Client Intercepted & passed on call to RunQueue function {function_name} with args:
		# 			#{args} and kwargs: {kwargs}"
		# 	),
		# 	aes_cypher_key= self._aes_session_key
		# )

		with self._method_reponse_dict_lock:
			self._method_id_counter += 1 #Increment the method id counter
			self._method_response_dict[self._method_id_counter] = queue.Queue(maxsize=1) #This makes it easy to wait
				# for the result of a method call - note that only 1 response is possible so queue size is 1
			method_call_id = self._method_id_counter

		log.debug(f"Sending intercepted call to RunQueue function {function_name} with args: {args} and kwargs: \
	    	{kwargs} with id {method_call_id} to server.")
		Transmission.send(
			self._socket,
			PickleTransmissionData(
				(
					PickledDataType.METHOD_CALL, #Indicates a triplet of the form(request_id, function_name, args, kwargs)
					(
						method_call_id,
						function_name,
						args,
						kwargs
					)
				)
			),
			aes_cypher_key= self._aes_session_key
		)
		try:
			if function_name == self.get_command_line_output.__name__: #TODO: might be a more elegant way to do this, maybe
				# request data in multiple chunks? Probably best to limit console-mirror-size as >50mb files
				# are not very UI-friendly
				return self._await_method_response(method_call_id, timeout=5) #Since log file can be quite large, wait longer

			return self._await_method_response(method_call_id, timeout=3) #Wait for the server to respond to the method call
				#and return the result
		except TimeoutError as exception:
			log.error(f"Timeout while waiting for response of {function_name}(...) with id {method_call_id} - {exception}")
			# raise exception
			return None

	def _await_method_response(self, function_response_id : int, timeout : float):
		"""
		Wait for server response after a function call. Response is automatically put in reponse_dict by the server-
		listener. If, after timeout, no response is received in this dict, a TimeoutError is raised.
		"""
		return_queue = self._method_response_dict[function_response_id]
		try:
			ret = return_queue.get(block=True, timeout=timeout)
			log.info(f"Received response from server for method call with id {function_response_id} of type {type(ret)}")
		except (TimeoutError, queue.Empty) as exception: #If the queue is empty, then the server did not respond in time #pylint: unused-variable
			raise TimeoutError(f"Timeout while waiting ({timeout}s) for response of method call with" 
		    	f"id {function_response_id}") #from exception
		finally: #Always clean up the queue to no longer listen for this id
			with self._method_reponse_dict_lock:
				del self._method_response_dict[function_response_id]

		if isinstance(ret, Exception): #Do not return exception types -> raise them instead #TODO: is this what we want?
			ret.args = (f"Exception raised by server while executing method call with id {function_response_id}: \
	       		{ret.args[0]}",) + ret.args[1:]
		return ret


	def get_connection_info(self):
		"""
		Returns a tuple of (server_ip, server_port, server_password) if connected, else tuple of (None, None, None)
		"""
		if self._socket is None:
			return (None, None, None)
		return (self._server_ip, self._server_port, self._password) #TODO: hashed password

	def disconnect_clean_server(self):
		"""
		Tries to disconnect socket if it is still active and resets all variables
		"""
		self._disconnect_flag = True
		self._server_ip = None
		self._server_port = None
		if self._socket is not None:
			try:
				self._socket.close()
			except OSError as exception:
				log.error(f"Error while closing socket: {exception}")
		self._socket = None
		self._aes_session_key = None
		if self._authenticated: #Only emit signal if we were connected before
			self.authenConnectionStateChanged.emit(False)
		self._authenticated = False
		self._connected_or_connecting = False
		if self._listen_thread is not None:
			#Wait for the listening thread to finish last iteration - then join it
			try:
				if self._listen_thread.is_alive():
					self._listen_thread.join()
			except Exception as exception: # pylint: disable=broad-exception-caught
				log.error(f"Error while joining listening thread when shutting down server: {exception}")
			self._listen_thread = None

		#TODO: stop threads

	def connect_to_server(self, server_ip: str, server_port: int, password : str) -> None:
		"""Try to connect to server with provided ip, port and password. If connection is successful, the client will
		start listening to the server for Transmissions.

		NOTE: this function will overwrite any existing connection to a server

		Args:
			server_ip (str): The server ip where RunQueueServer is running
			server_port (int): The server port where RunQueueServer is running
			password (str): The password to use to authenticate with the server
		"""
		try:
			self.disconnect_clean_server()
			self._connected_or_connecting = True
			self._authenticated = False
			self._disconnect_flag = False
			self._server_ip = server_ip
			self._server_port = server_port

			#Connect tcp socket to server
			self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self._socket.connect((self._server_ip, self._server_port))

			#Send public key to server
			Transmission.send(
				self._socket,
				PubKeyTransmissionData(
					self._public_key.export_key()
				)
			)


			received = Transmission.receive(
				self._socket,
				None
			)
			if received.transmission_type != TransmissionType.SESSION_KEY:
				log.error("Did not receive AES session key from server")
				#Check if remote server closed connection
				if received.transmission_type == TransmissionType.STATE:
					state = StateTransmissionData.from_transmission_bytes(received.transmission_data)
					log.error(f"Server returned: {state.state_msg_type.name} - {state.state_msg}")
					raise AuthenticationException(f"Server returned: {state.state_msg_type.name} - {state.state_msg}")
					#TODO: login-exception instead of Exception
				raise AuthenticationException("Did not receive AES session key from server during login process")

			#Decrypt the session key using the private key
			encrypted_aes_session_key = AESSessionKeyTransmissionData.from_transmission_bytes(received.transmission_data)
			aes_session_key = self._private_rsa_cipher.decrypt(encrypted_aes_session_key.encrypted_session_key)
			# self._aes_cipher = AES.new(aes_session_key, AES.MODE_EAX)

			log.info(f"Aes session key: {aes_session_key} - now verifying password")

			#=========== From here on out, all communication is encrypted ===========
			#Send login data
			Transmission.send(
				self._socket,
				LoginTransmissionData(
					password
				),
				aes_session_key
			)

			self._aes_session_key = aes_session_key
			#Wait for response
			received = Transmission.receive(
				self._socket,
				aes_session_key
			)
			if received.transmission_type != TransmissionType.STATE:
				log.error("Did not receive login response from server")
				raise AuthenticationException("Did not receive login response from server") #TODO: login-error

			state_msg = StateTransmissionData.from_transmission_bytes(received.transmission_data)
			if state_msg.state_msg_type == StateMsgType.LOGIN_ERROR:
				raise AuthenticationException(f"Login rejected by server: {state_msg.state_msg}")
			elif state_msg.state_msg_type == StateMsgType.LOGIN_ACCEPTED:
				log.info("Login accepted by server")




			if self._listen_thread is None:
				log.info("Starting server-listening thread...")
				self._listen_thread = threading.Thread(target=self._listen_to_server)
				self._listen_thread.start()
				log.info("Started server-listening thread...")

			self._authenticated = True
			self._password = password #TODO: hash password twice.
			self.authenConnectionStateChanged.emit(True)
			self.resetTriggered.emit() #Emit the reset signal so all models re-request all data
			log.info("Connected to server and authenticated. Can now request/receive/transmit data from/to server.")
		except Exception as exception:
			#Just pass on the exception
			self.disconnect_clean_server() #Clean up
			raise exception


	def _listen_to_server(self):
		"""
		Start listening to the server for Transmissions. E.g.:
		- StateTransmissionData -> Server is sending a state message
		- PickleTransmissionData -> Server is sending a pickled object - for example - a result of a function call or
			a Signal object
		"""
		log.info("Started continuously listening to server")
		try:
			while not self._disconnect_flag:
				received = Transmission.receive(
					self._socket, #type: ignore
					aes_cipher_key=self._aes_session_key
				)

				if received.transmission_type == TransmissionType.STATE:
					data = StateTransmissionData.from_transmission_bytes(received.transmission_data)
					log.info(
						f"Received state from server: {data.state_msg_type.name} - {data.state_msg}"
					)
				elif received.transmission_type == TransmissionType.PICKLED_OBJECT:
					try:
						data = PickleTransmissionData.from_transmission_bytes(received.transmission_data)
						unpickled_data : tuple = data.unpickled_data #type: ignore
						assert isinstance(unpickled_data, tuple), ("First element is the type of the data,"
							" should be tuple")
						unpickled_data = tuple(unpickled_data) #For type hinting
						# unpickled_data : tuple = unpickled_data

						if unpickled_data[0] == PickledDataType.METHOD_RETURN:
							#The server is sending the result of a function call
							#Unpack the data
							function_call_id, result = unpickled_data[1]
							with self._method_reponse_dict_lock:
								if function_call_id not in self._method_response_dict:
									log.error(f"Received function call result with id {function_call_id} - but no "
										"function call with that id was made")
									continue
								self._method_response_dict[function_call_id].put(result, block=True)
						elif unpickled_data[0] == PickledDataType.SIGNAL_EMIT:
							#The server is sending a signal
							#Unpack the data
							signal_name, args = unpickled_data[1]

							#Get signal by name
							signal : PySignal.Signal = getattr(self, signal_name) #NOT QT SIGNAL (actually classignal)
							log.debug(
								f"Received signal {signal_name} from server - resulting in a re-transmit of signal"
									f"{signal} of type {type(signal)}"
							)
							assert isinstance(signal, PySignal.Signal)
							#Emit the signal
							signal.emit(*args) #type: ignore
						else:
							log.error(f"Received pickled data of type {unpickled_data[0].name} - which is not "
		 						"supported (at the client side)")
					except Exception as exception: # pylint: disable=broad-exception-caught
						log.error(f"Error while unpickling pickle-transmission from server - {type(exception)}: "
							f"{exception}")
						traceback.print_exc()

		except Exception as exception: # pylint: disable=broad-exception-caught
			log.error(f"Error while listening to server: {exception} - disconnecting")
			self.disconnect_clean_server()
			return



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'run_queue_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.5.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QHeaderView, QPushButton,
    QSizePolicy, QSpacerItem, QToolButton, QVBoxLayout,
    QWidget)

from configurun.app.views.run_queue_tree_view import RunQueueTreeView
from pyside6_utils.widgets.square_frame import SquareFrame
import configurun.res.app_resources_rc

class Ui_RunQueueWidget(object):
    def setupUi(self, RunQueueWidget):
        if not RunQueueWidget.objectName():
            RunQueueWidget.setObjectName(u"RunQueueWidget")
        RunQueueWidget.resize(596, 551)
        self.verticalLayout = QVBoxLayout(RunQueueWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, -1, 0, -1)
        self.squareFrame_3 = SquareFrame(RunQueueWidget)
        self.squareFrame_3.setObjectName(u"squareFrame_3")
        self.verticalLayout_5 = QVBoxLayout(self.squareFrame_3)
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.MoveUpInQueueBtn = QPushButton(self.squareFrame_3)
        self.MoveUpInQueueBtn.setObjectName(u"MoveUpInQueueBtn")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.MoveUpInQueueBtn.sizePolicy().hasHeightForWidth())
        self.MoveUpInQueueBtn.setSizePolicy(sizePolicy)
        self.MoveUpInQueueBtn.setMinimumSize(QSize(50, 50))
        self.MoveUpInQueueBtn.setSizeIncrement(QSize(1, 1))
        icon = QIcon()
        icon.addFile(u":/Icons/icons/actions/go-up.png", QSize(), QIcon.Normal, QIcon.Off)
        self.MoveUpInQueueBtn.setIcon(icon)
        self.MoveUpInQueueBtn.setIconSize(QSize(25, 25))

        self.verticalLayout_5.addWidget(self.MoveUpInQueueBtn)


        self.horizontalLayout.addWidget(self.squareFrame_3)

        self.squareFrame_4 = SquareFrame(RunQueueWidget)
        self.squareFrame_4.setObjectName(u"squareFrame_4")
        self.verticalLayout_6 = QVBoxLayout(self.squareFrame_4)
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.MoveDownInQueueBtn = QPushButton(self.squareFrame_4)
        self.MoveDownInQueueBtn.setObjectName(u"MoveDownInQueueBtn")
        sizePolicy.setHeightForWidth(self.MoveDownInQueueBtn.sizePolicy().hasHeightForWidth())
        self.MoveDownInQueueBtn.setSizePolicy(sizePolicy)
        self.MoveDownInQueueBtn.setMinimumSize(QSize(50, 50))
        self.MoveDownInQueueBtn.setSizeIncrement(QSize(1, 1))
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/actions/go-down.png", QSize(), QIcon.Normal, QIcon.Off)
        self.MoveDownInQueueBtn.setIcon(icon1)
        self.MoveDownInQueueBtn.setIconSize(QSize(25, 25))

        self.verticalLayout_6.addWidget(self.MoveDownInQueueBtn)


        self.horizontalLayout.addWidget(self.squareFrame_4)

        self.squareFrame_5 = SquareFrame(RunQueueWidget)
        self.squareFrame_5.setObjectName(u"squareFrame_5")
        self.verticalLayout_7 = QVBoxLayout(self.squareFrame_5)
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.CancelStopButton = QPushButton(self.squareFrame_5)
        self.CancelStopButton.setObjectName(u"CancelStopButton")
        sizePolicy.setHeightForWidth(self.CancelStopButton.sizePolicy().hasHeightForWidth())
        self.CancelStopButton.setSizePolicy(sizePolicy)
        self.CancelStopButton.setMinimumSize(QSize(50, 50))
        self.CancelStopButton.setSizeIncrement(QSize(1, 1))
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/actions/process-stop.png", QSize(), QIcon.Normal, QIcon.Off)
        self.CancelStopButton.setIcon(icon2)
        self.CancelStopButton.setIconSize(QSize(25, 25))

        self.verticalLayout_7.addWidget(self.CancelStopButton)


        self.horizontalLayout.addWidget(self.squareFrame_5)

        self.squareFrame_6 = SquareFrame(RunQueueWidget)
        self.squareFrame_6.setObjectName(u"squareFrame_6")
        self.verticalLayout_8 = QVBoxLayout(self.squareFrame_6)
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.verticalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.DeleteButton = QPushButton(self.squareFrame_6)
        self.DeleteButton.setObjectName(u"DeleteButton")
        sizePolicy.setHeightForWidth(self.DeleteButton.sizePolicy().hasHeightForWidth())
        self.DeleteButton.setSizePolicy(sizePolicy)
        self.DeleteButton.setMinimumSize(QSize(50, 50))
        self.DeleteButton.setSizeIncrement(QSize(1, 1))
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/places/user-trash.png", QSize(), QIcon.Normal, QIcon.Off)
        self.DeleteButton.setIcon(icon3)
        self.DeleteButton.setIconSize(QSize(25, 25))

        self.verticalLayout_8.addWidget(self.DeleteButton)


        self.horizontalLayout.addWidget(self.squareFrame_6)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.squareFrame = SquareFrame(RunQueueWidget)
        self.squareFrame.setObjectName(u"squareFrame")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.squareFrame.sizePolicy().hasHeightForWidth())
        self.squareFrame.setSizePolicy(sizePolicy1)
        self.verticalLayout_3 = QVBoxLayout(self.squareFrame)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.StartRunningQueueBtn = QPushButton(self.squareFrame)
        self.StartRunningQueueBtn.setObjectName(u"StartRunningQueueBtn")
        sizePolicy2 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.StartRunningQueueBtn.sizePolicy().hasHeightForWidth())
        self.StartRunningQueueBtn.setSizePolicy(sizePolicy2)
        self.StartRunningQueueBtn.setMinimumSize(QSize(50, 50))
        icon4 = QIcon()
        icon4.addFile(u":/Icons/icons/actions/media-playback-start.png", QSize(), QIcon.Normal, QIcon.Off)
        icon4.addFile(u":/Icons/icons/media-playback-started.png", QSize(), QIcon.Normal, QIcon.On)
        self.StartRunningQueueBtn.setIcon(icon4)
        self.StartRunningQueueBtn.setIconSize(QSize(25, 25))
        self.StartRunningQueueBtn.setCheckable(True)

        self.verticalLayout_3.addWidget(self.StartRunningQueueBtn)


        self.horizontalLayout.addWidget(self.squareFrame)

        self.toolButton = QToolButton(RunQueueWidget)
        self.toolButton.setObjectName(u"toolButton")
        sizePolicy3 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.toolButton.sizePolicy().hasHeightForWidth())
        self.toolButton.setSizePolicy(sizePolicy3)
        self.toolButton.setMinimumSize(QSize(35, 35))
        icon5 = QIcon()
        icon5.addFile(u":/Icons/icons/categories/applications-system.png", QSize(), QIcon.Normal, QIcon.Off)
        self.toolButton.setIcon(icon5)
        self.toolButton.setIconSize(QSize(25, 25))

        self.horizontalLayout.addWidget(self.toolButton)

        self.squareFrame_2 = SquareFrame(RunQueueWidget)
        self.squareFrame_2.setObjectName(u"squareFrame_2")
        self.verticalLayout_4 = QVBoxLayout(self.squareFrame_2)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")

        self.horizontalLayout.addWidget(self.squareFrame_2)


        self.horizontalLayout_2.addLayout(self.horizontalLayout)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.QueueViewLayout = QVBoxLayout()
        self.QueueViewLayout.setObjectName(u"QueueViewLayout")
        self.runQueueTreeView = RunQueueTreeView(RunQueueWidget)
        self.runQueueTreeView.setObjectName(u"runQueueTreeView")

        self.QueueViewLayout.addWidget(self.runQueueTreeView)


        self.verticalLayout.addLayout(self.QueueViewLayout)

        self.verticalLayout.setStretch(1, 1)

        self.retranslateUi(RunQueueWidget)

        QMetaObject.connectSlotsByName(RunQueueWidget)
    # setupUi

    def retranslateUi(self, RunQueueWidget):
        RunQueueWidget.setWindowTitle(QCoreApplication.translate("RunQueueWidget", u"Form", None))
#if QT_CONFIG(tooltip)
        self.MoveUpInQueueBtn.setToolTip(QCoreApplication.translate("RunQueueWidget", u"Move item up in Queue", None))
#endif // QT_CONFIG(tooltip)
        self.MoveUpInQueueBtn.setText("")
#if QT_CONFIG(tooltip)
        self.MoveDownInQueueBtn.setToolTip(QCoreApplication.translate("RunQueueWidget", u"Move item down in Queue", None))
#endif // QT_CONFIG(tooltip)
        self.MoveDownInQueueBtn.setText("")
#if QT_CONFIG(tooltip)
        self.CancelStopButton.setToolTip(QCoreApplication.translate("RunQueueWidget", u"Dequeue/Stop item", None))
#endif // QT_CONFIG(tooltip)
        self.CancelStopButton.setText("")
#if QT_CONFIG(tooltip)
        self.DeleteButton.setToolTip(QCoreApplication.translate("RunQueueWidget", u"Delete item", None))
#endif // QT_CONFIG(tooltip)
        self.DeleteButton.setText("")
#if QT_CONFIG(tooltip)
        self.StartRunningQueueBtn.setToolTip(QCoreApplication.translate("RunQueueWidget", u"Start automatic run-mode", None))
#endif // QT_CONFIG(tooltip)
        self.StartRunningQueueBtn.setText("")
        self.toolButton.setText(QCoreApplication.translate("RunQueueWidget", u"...", None))
    # retranslateUi




"""
This script runs the example in which we can select our model/dataset type.
"""



import logging
import typing

from configurun.configuration.base_options import BaseOptions
from configurun.configuration.configuration import Configuration
from configurun.examples.example_options.example_options import (
    ExampleDatasetOptions, ExampleGeneralOptions, ExampleMainOptions,
    ExampleModelOptions, ExtendedExampleDatasetOptions,
    ExtendedExampleModelOptions)
from configurun.examples.example_target_function import example_target_function

log = logging.getLogger(__name__)

def example_deduce_new_option_classes(
			#NOTE: all suboptions of the configucation can be accesed via the configuration.<attr> syntax.
			# For type hinting, we can create a union of all the suboptions classes we would like to use
			# in this example, we only use the main options. The actual type of the configuration argument is
			# Configuration, but we can use the ExampleMainOptions class as a type hint
			configuration : Configuration #|ExampleMainOptions | ExampleDatasetOptions | ExampleGeneralOptions #type: ignore
		) -> typing.Dict[str, typing.Type[BaseOptions] | typing.Type[None]]:
	"""Deduce the new option classes from the current configuration

	Based on the selection of the mainoptions, we can deduce the new option class types, in this example:
		- main_options : Always ExampleMainOptions
		- general_options : Always ExampleGeneralOptions
		- model_options : ExampleDatasetOptions | ExtendedExampleDatasetOptions
		- dataset_options : ExampleDatasetOptions | ExtendedExampleDatasetOptions

	Args:
		configuration (Configuration): The current configuration, we can typehint the main options
	"""
	configuration : ExampleMainOptions = configuration #Typehint the configuration  #type: ignore

	ret_dict = { #Always return the same main and general options
		"main_options" : ExampleMainOptions,
		"general_options" : ExampleGeneralOptions
	}

	#========== Select the dataset-class based on the dataset-selection in main-options ==========
	if hasattr(configuration, "dataset_type"):
		if configuration.dataset_type is not None and configuration.dataset_type == "ExampleDataset":
			ret_dict["dataset_options"] = ExampleDatasetOptions
		elif configuration.dataset_type is not None and configuration.dataset_type == "ExtendedExampleDataset":
			ret_dict["dataset_options"] = ExtendedExampleDatasetOptions
		else:
			# ret_dict["dataset_options"] = None #No dataset options - uncomment to display empty window
			pass #Remove if None selected
	# else: #NOTE: if we want to display an empty options-window, we can set the dataset_options to None
	# 	ret_dict["dataset_options"] = None



	#=========== Select the model based on the model-selection in main-options ===========
	if hasattr(configuration, "model_type"):
		if configuration.model_type is not None and configuration.model_type == "ExampleModel":
			ret_dict["model_options"] = ExampleModelOptions
		elif configuration.model_type is not None and configuration.model_type == "ExtendedExampleModel":
			ret_dict["model_options"] = ExtendedExampleModelOptions
		else:
			# ret_dict["model_options"] = None #No Model options Uncomment to display empty window
			pass #Remove if None selected
	# else: #NOTE: if we want to display an empty model-options-window, we can set the dataset_options to None
	# 	ret_dict["model_options"] = None

	return ret_dict


def run_example_app(log_level=logging.INFO):
	"""Run an example instance of the Configurun-App using the example run function and example-option-deducer
	"""
	# pylint: disable=import-outside-toplevel
	from configurun.app.create_run import \
	    run_local
	import os
	from configurun.app.main_window import APP_NAME
	import tempfile

	#=========================== Create the app using the example ===========================
	tempdir = tempfile.gettempdir()
	workspace_path = os.path.join(tempdir, APP_NAME, "Configurun-Local-Example")
	log.info(f"Saving example app workspace to {workspace_path}")
	run_local(
		target_function=example_target_function,
		options_source=example_deduce_new_option_classes,
		workspace_path=workspace_path,
		log_level=log_level
	)



if __name__ == "__main__":
	run_example_app(logging.DEBUG)



"""
Implements a automatic/dynamic dataclass generation for several sklearn models.
"""
# from MachineLearning.framework.options.mvts_options import MVTSGeneralOptions
import inspect
import logging
import sys
import typing
from dataclasses import dataclass, field, make_dataclass

# import dill
from numpydoc.docscrape import NumpyDocString
from pyside6_utils.classes.constraints import StrOptions
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
# from sklearn.gaussian_process.kernels import RBF
from sklearn.linear_model import RidgeClassifierCV
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from configurun.configuration.base_options import BaseOptions


class SklearnClass:
	"""Example class for sklearn model - all sklearn classes have a parameter constraints dict."""
	_parameter_constraints: dict = {}


log = logging.getLogger(__name__)
_NAME_ALGORITHM_DICT : dict[str, type[SklearnClass | typing.Any]] = {
							"SKLEARN_NearestNeighbors" : KNeighborsClassifier,
							"SKLEARN_LinearSVM" : SVC,
							"SKLEARN_GaussianProcessClassifier" : GaussianProcessClassifier,
							"SKLEARN_DecisionTreeClassifier" : DecisionTreeClassifier,
							"SKLEARN_RandomForestClassifier" : RandomForestClassifier,
							"SKLEARN_MPLClassifier" : MLPClassifier,
							"SKLEARN_AdaBoostClassifier" : AdaBoostClassifier,
							"SKLEARN_NaiveBayes" : GaussianNB,
							"SKLEARN_RidgeClassifierCV" : RidgeClassifierCV
							#Add to this list to add more algorithms
}

_NAME_DOC_DICT : dict[str, NumpyDocString] = { #This dictionary will hold all the docstrings for each algorithm
	name : NumpyDocString(algorithm.__doc__) for name, algorithm in _NAME_ALGORITHM_DICT.items()
}

_NAME_SIGNATUR_DICT : dict[str, inspect.Signature] = { #This dictionary will hold all the signatures for each algorithm
	name : inspect.signature(algorithm) for name, algorithm in _NAME_ALGORITHM_DICT.items()
}

NAME_COMMENTS_DICT : dict[str, str] = { #This dictionary will hold all the comments for each algorithm
	key: algorithm.__doc__ if algorithm.__doc__ is not None else "No comments available"\
		for key, algorithm in _NAME_ALGORITHM_DICT.items()
}



def _get_model_options(algorithm_name : str) -> type:
	"""Sklearn classes all have a constraints dictionary for the types and limits of each parameter. We use this
	to automatically generate a dataclass for each algorithm which can be used to create a UI for each algorithm.
	Adding more algorithms should then be as simple as adding the algorithm to the _name_algorithm_dict dictionary.

	Args:
		algorithm_name (str): Name of the algorithm (key in _name_algorithm_dict)

	Returns:
		type: A dataclass type with all settings of the algorithm as fields
	"""
	global _NAME_DOC_DICT, _NAME_SIGNATUR_DICT, _NAME_ALGORITHM_DICT #pylint: disable=global-variable-not-assigned
	doc = _NAME_DOC_DICT[algorithm_name]
	algorithm : type[SklearnClass] = _NAME_ALGORITHM_DICT[algorithm_name]
	parameters = doc['Parameters'] #Parameters of the algorithm
	parsed_description_dict = {
		arg.name : "\n".join(arg.desc) for arg in parameters
	}
	function_signature = _NAME_SIGNATUR_DICT[algorithm_name] #Get the signature of the algorithm class using the doc
	args = []
	for param_name, param_data in function_signature.parameters.items():
		if param_name == 'self':
			continue
		args.append(( #Create a tuple: (parameter name, type, field(<neccesary data for UI>))
						param_name, #Name of the parameter
						#Try to get the type of the parameter, if not possible, use Any:
						type(param_data.default) if param_data.default is not inspect.Parameter.empty else any,
						field(default=param_data.default if param_data.default is not inspect.Parameter.empty else None,
								metadata=dict(
										#Replace underscores with spaces and capitalize each word
										display_name = param_name.replace('_', ' ').title(),
										#Get the help string from the docstring:
										help=parsed_description_dict[param_name] \
											if param_name in parsed_description_dict else None,
										#Get the constraints from the docstring:
										constraints=algorithm._parameter_constraints[param_name] #pylint: disable=protected-access\
											if param_name in algorithm._parameter_constraints else None #pylint: disable=protected-access
								)

							)

					)
		)

	new_dataclass = make_dataclass(f"{algorithm_name}ModelOptions",
									args,
									bases=(BaseOptions,)
									# namespace = { "__reduce__"} #NOTE: We can also use this to make class pickleable
									)

	return new_dataclass

SKLEARN_NAME_DATACLASS_DICT = { #This dictionary will hold all the dataclasses for the options of each algorithm

}

for key, val in _NAME_ALGORITHM_DICT.items():
	class_def = _get_model_options(key)
	#make_dataclass does not set the module attribute of the new class appropriately (as of python 3.10.8),
	# so we have to do it manually
	try: #NOTE: this should be fixed in Python version 3.12.0 alpha 7
			# (https://github.com/python/cpython/commit/b48be8fa18518583abb21bf6e4f5d7e4b5c9d7b2)
			#make_dataclass now has a module parameter which fixes this issue (even if not specified)
		module = sys._getframemodulename(0) or '__main__' #pylint: disable=protected-access #type:ignore
	except AttributeError:
		try:
			module = sys._getframe(0).f_globals.get('__name__', '__main__') #pylint: disable=protected-access
		except (AttributeError, ValueError):
			log.error(f"Could not set module attribute of dataclass {class_def.__name__}")
			continue
	class_def.__module__ = module #Set the module attribute of the new class to the module of the caller
	SKLEARN_NAME_DATACLASS_DICT[key] = class_def
	globals()[class_def.__name__] = class_def #Add class to global scope to enable pickling


@dataclass
class SklearnMainOptions(BaseOptions):
	"""
	Similar to example options, but with the choice of selecting a sklearn model
	"""

	# def __post_init__(self):
	# 	name_field_dict = {field.name: field for field in fields(self)}
		# for algorithm in SklearnModelOptions.get_name_algorithm_dict().keys():
		# 		#Make sure we added all possible algorithms to the model options (inside .Literal)
		# 	assert algorithm in name_field_dict["model"].type.__args__,
		# 		f"Algorithm {algorithm} not in {name_field_dict['model'].type.__args__}"

	# Model
	model_type : typing.Literal[" "] | None = field( #We dynamically set the constraints of this field using the
			# 'constraints' parameter, so we can apply the dynamically generated sklearn-model
		default=None,
		metadata=dict(
					#TODO: add a required flag
			display_name="Model",
			constraints_help= {
					"MVTSMODEL" : ("A model especially intended in use for time series representation"
					 	"learning based on the transformer encoder architecture"),
					"LINEAR":  "A simple baseline Linear Model"} #Take union of other method key-dict and this one
				| { name : comment for name, comment in NAME_COMMENTS_DICT.items()}, #Get doc-description of each algo
			help=("What model class to use" #Add the (short description) of each model to the help string
				),
			constraints = [StrOptions({ "MVTSMODEL", "LINEAR", *(NAME_COMMENTS_DICT.keys())})]
		)
	)

	#Dataset Type
	dataset_type : typing.Literal['exampleDataclass', "extendedExampleDataclass"] | None = field(
		default=None,
		metadata=dict(
			display_name="Dataset Type",
			help="""Which type of dataset is going to be processed."""
		)
	)



if __name__ == "__main__":
	print("Running SklearnModelOptions.py")
	for cur_name, cur_class in SKLEARN_NAME_DATACLASS_DICT.items():
		print(f"{cur_name} : {cur_class}")
	print("Done")
